from cilantro_ee.utils.test.testnet_config import set_testnet_config
set_testnet_config('2-2-2.json')
#
from cilantro_ee.core.logger.base import get_logger
# from cilantro_ee.constants.system_config import *
from cilantro_ee.utils.utils import int_to_bytes, bytes_to_int
from cilantro_ee.utils.hasher import Hasher
#
from contracting.db.cr.client import SubBlockClient
from contracting.db.cr.callback_data import ExecutionData, SBData
#
# from cilantro_ee.nodes.delegate.block_manager import IPC_IP, IPC_PORT
from cilantro_ee.nodes.delegate.sub_block_builder import SubBlockBuilder, SubBlockManager, NextBlockState
#
from cilantro_ee.core.messages.message_type import MessageType
from cilantro_ee.core.messages.message import Message
#
from unittest import TestCase
from unittest import mock
from unittest.mock import MagicMock
from cilantro_ee.services.storage.vkbook import PhoneBook, VKBook
from cilantro_ee.constants.testnet import *
from cilantro_ee.services.storage.vkbook import PhoneBook
import asyncio
import time
from cilantro_ee.constants.testnet import TESTNET_DELEGATES, TESTNET_MASTERNODES
from cilantro_ee.core.crypto import wallet
#
_log = get_logger("TestSubBlockBuilder")
#
TEST_IP = '127.0.0.1'
# DELEGATE_SK = TESTNET_DELEGATES[0]['sk']
DELEGATE_SK = PhoneBook.delegates[0]

IPC_IP = 'block-manager-ipc-sock'
IPC_PORT = 6967
#
# MN_SK1 = TESTNET_MASTERNODES[0]['sk']
# MN_SK2 = TESTNET_MASTERNODES[1]['sk']
MN_SK1 = PhoneBook.masternodes[0]
MN_SK2 = PhoneBook.masternodes[1]
MN_SK1 = TESTNET_MASTERNODES[0]['sk']
MN_SK2 = TESTNET_MASTERNODES[1]['sk']

#
#
class SBBTester:

    @staticmethod
    def test(func):
        @mock.patch("cilantro_ee.core.utils.worker.asyncio")
        @mock.patch("cilantro_ee.core.utils.worker.SocketManager")
        @mock.patch("cilantro_ee.nodes.delegate.block_manager.asyncio")
        @mock.patch("cilantro_ee.nodes.delegate.block_manager.SubBlockBuilder.run")
        def _func(*args, **kwargs):
            return func(*args, **kwargs)
        return _func

    @staticmethod
    def send_ipc_to_sbb(sbb: SubBlockBuilder, msg: bytes):
        frames = [int_to_bytes(0), msg]
        sbb.handle_ipc_msg(frames)

    @staticmethod
    def send_sub_to_sbb(sbb: SubBlockBuilder, envelope, handler_key, filter_frame=b''):
        frames = [filter_frame, int_to_bytes(0), b'']
        sbb.handle_sub_msg(frames, handler_key)

#    @staticmethod
#    def create_tx_batch_env(num_txs: int, env_signing_key: str) -> Envelope:
#        tx_batch = build_test_transaction_batch(num_tx=num_txs)
#        return Envelope.create_from_message(message=tx_batch, signing_key=env_signing_key)
#

class TestSubBlockBuilder(TestCase):

    @SBBTester.test
    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_SB_PER_BUILDER", 8)
    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_SB_PER_BLOCK_PER_BUILDER", 1)
    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_SUB_BLOCKS", 8)
    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_SB_BUILDERS", 1)
    def test_create_sub_sockets(self, *args):
        sbb = SubBlockBuilder(ip=TEST_IP, signing_key=DELEGATE_SK, sbb_index=0, ipc_ip=IPC_IP, ipc_port=IPC_PORT)

        self.assertEqual(len(sbb.sb_managers), 8)

    def run_async(self, loop, period=1.0):
        async def asyncio_sucks():
            await asyncio.sleep(period)

        loop.run_until_complete(asyncio_sucks())

    def run_loops(self, *loops, period=1.0):
        async def asyncio_sucks():
            await asyncio.sleep(period)

        master_loop = asyncio.new_event_loop()
        asyncio.set_event_loop(master_loop)
        futs = []
        for l in loops:
            futs.append(asyncio.ensure_future(l.run_until_complete(asyncio_sucks())))

        master_loop.run_until_complete(asyncio.gather(futs))



    @SBBTester.test
    def test_create_empty_sbc(self, *args):
        sbb = SubBlockBuilder(ip=TEST_IP, signing_key=DELEGATE_SK, sbb_index=0, ipc_ip=IPC_IP, ipc_port=IPC_PORT)

        PhoneBook = VKBook(delegates=[sbb.verifying_key], masternodes=[TESTNET_MASTERNODES[0]['vk']])

        input_hash = 'A' * 64
        sbb_idx = 0
        sb_data = SBData(input_hash, [])
        sbb._send_msg_over_ipc = MagicMock()

        sbb._create_empty_sbc(sb_data)
        sbb._send_msg_over_ipc.assert_called()
        empty_sbc = sbb._send_msg_over_ipc.call_args[0][0]
        self.assertTrue(isinstance(empty_sbc, SubBlockContender))
        self.assertEqual(empty_sbc.input_hash, input_hash)
        self.assertEqual(empty_sbc.sb_index, sbb_idx)
        self.assertTrue(empty_sbc.is_empty)


#    @SBBTester.test
#    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_SUB_BLOCKS", 1)
#    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_BLOCKS", 1)
#    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_SB_BUILDERS", 1)
#    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_SB_PER_BUILDER", 1)
#    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_SB_PER_BLOCK_PER_BUILDER", 1)
#    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.TRANSACTIONS_PER_SUB_BLOCK", 4)
#    def test_one_sb_recv_empty_tx_bag(self, *args):
#        sbb = SubBlockBuilder(ip=TEST_IP, signing_key=DELEGATE_SK, sbb_index=0, ipc_ip=IPC_IP, ipc_port=IPC_PORT)
#        self.assertEqual(len(sbb.sb_managers), 1)
#
#        # Send an empty TX batch. This should get queued up as the SBB doesnt have a MakeNextBlock msg yet
#        empty_tx_batch_env = SBBTester.create_tx_batch_env(num_txs=0, env_signing_key=MN_SK1)
#        SBBTester.send_sub_to_sbb(sbb, envelope=empty_tx_batch_env, handler_key=0)
#        self.assertEqual(len(sbb.sb_managers[0].pending_txs), 1)
#
#        # Now, send a MakeNextBlock notif. This should trigger an empty subblock to be built and sent over IPC to BM
#        make_next_block = MakeNextBlock.create()
#        SBBTester.send_ipc_to_sbb(sbb, make_next_block)
#
#    @SBBTester.test
#    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_SUB_BLOCKS", 1)
#    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_BLOCKS", 1)
#    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_SB_BUILDERS", 1)
#    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_SB_PER_BLOCK", 1)
#    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_SB_PER_BUILDER", 1)
#    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_SB_PER_BLOCK_PER_BUILDER", 1)
#    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.TRANSACTIONS_PER_SUB_BLOCK", 4)
#    def test_sub_msg_with_make_next_block_notification_calls_handle_ipc_msg(self, *args):
#        """
#        Tests handle_ipc_msg correctly calls handle_new_block when a NewBlockNotification is received
#        """
#
#        sbb = SubBlockBuilder(ip=TEST_IP, signing_key=DELEGATE_SK, sbb_index=0, ipc_ip=IPC_IP, ipc_port=IPC_PORT)
#
#        PhoneBook = VKBook(delegates=[sbb.verifying_key], masternodes=[TESTNET_MASTERNODES[0]['vk']])
#
#        sbb._send_msg_over_ipc = MagicMock()
#
#        self.assertTrue(len(sbb.sb_managers) == 1)
#
#        # Mock Envelope.from_bytes to return a mock envelope of our choosing
#        tx_batch_env = SBBTester.create_tx_batch_env(num_txs=4, env_signing_key=MN_SK1)
#        # print("print tsns")
#        # print(tx_batch_env.message.transactions[0])
#        SBBTester.send_sub_to_sbb(sbb, envelope=tx_batch_env, handler_key=0)
#
#
#        sbb._make_next_sub_block()
#
#        self.run_async(sbb.loop, 2)
#        assert sbb._send_msg_over_ipc.called
#        assert sbb._send_msg_over_ipc.call_count == 2
#        msg = sbb._send_msg_over_ipc.call_args[1][0][0]
#        print("raghu msg = {}".format(msg))
#        self.assertTrue(isinstance(msg, SubBlockContender))
#        self.assertFalse(msg.is_empty)


    @SBBTester.test
    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_SUB_BLOCKS", 1)
    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_BLOCKS", 1)
    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_SB_BUILDERS", 1)
    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_SB_PER_BUILDER", 1)
    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_SB_PER_BLOCK_PER_BUILDER", 1)
    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.TRANSACTIONS_PER_SUB_BLOCK", 4)
    def test_handle_new_block_signal_calls_make_next_sub_block(self, *args):
        sbb = SubBlockBuilder(ip=TEST_IP, signing_key=DELEGATE_SK, sbb_index=0, ipc_ip=IPC_IP, ipc_port=IPC_PORT)
        sbb._make_next_sub_block = MagicMock()

        make_next_block = MakeNextBlock.create()
        SBBTester.send_ipc_to_sbb(sbb, make_next_block)

        sbb._make_next_sub_block.assert_called_once()


    @SBBTester.test
    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_SUB_BLOCKS", 1)
    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_BLOCKS", 1)
    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_SB_BUILDERS", 1)
    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_SB_PER_BUILDER", 1)
    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_SB_PER_BLOCK", 1)
    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_SB_PER_BLOCK_PER_BUILDER", 1)
    # @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.TRANSACTIONS_PER_SUB_BLOCK", 4)
    def test_align_input_hash_signal(self, *args):
        TRANSACTIONS_PER_SUB_BLOCK = 4
        sbb = SubBlockBuilder(ip=TEST_IP, signing_key=DELEGATE_SK, sbb_index=0, ipc_ip=IPC_IP, ipc_port=IPC_PORT)



        self.assertEqual(len(sbb.sb_managers), 1)
        sbb._send_msg_over_ipc = MagicMock()
        sbb._create_empty_sbc = MagicMock()
        sbb._create_sbc_from_batch = MagicMock()

        # send it 4 blocks w/ 4 txns
        ib_hashes = []
        num_txs = TRANSACTIONS_PER_SUB_BLOCK
        for i in range(4):
            tx_batch = SBBTester.create_tx_batch_env(num_txs=num_txs, env_signing_key=MN_SK1)
            input_hash = Hasher.hash(tx_batch)
            ib_hashes.append(input_hash)
            SBBTester.send_sub_to_sbb(sbb, envelope=tx_batch, handler_key=0)

        # Now, send a MakeNextBlock notif. This should trigger an empty subblock to be built and sent over IPC to BM
        make_next_block = MakeNextBlock.create()
        SBBTester.send_ipc_to_sbb(sbb, make_next_block)

        self.run_async(sbb.loop, 2)
        # should receive a _new subblock for input 0 and assert its input_hash is same as the one above
        # SBBTester.send_ipc_to_sbb(sbb, make_next_block)
        sbb._create_sbc_from_batch.assert_called_once()
        sb = sbb._create_sbc_from_batch.call_args[0][0]
        self.assertEqual(sb.input_hash, ib_hashes[0])

        # need to update the test case here
        # send align notification for input hash # 3
        # message = AlignInputHash.create(ib_hashes[2], 0)
        # SBBTester.send_ipc_to_sbb(sbb, message)
        # SBBTester.send_ipc_to_sbb(sbb, make_next_block)
        # self.run_async(sbb.loop, 2)
        # should receive a _new subblock for input 4
        # sbb._create_sbc_from_batch.assert_called()
        # sb = sbb._create_sbc_from_batch.call_args[0][0]
        # self.assertEqual(sb.input_hash, ib_hashes[3])


    @SBBTester.test
    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_SUB_BLOCKS", 1)
    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_BLOCKS", 1)
    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_SB_BUILDERS", 1)
    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_SB_PER_BUILDER", 1)
    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_SB_PER_BLOCK_PER_BUILDER", 1)
    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.TRANSACTIONS_PER_SUB_BLOCK", 4)
    def test_single_tx_batch_adds_to_queue(self, *args):
        sbb = SubBlockBuilder(ip=TEST_IP, signing_key=DELEGATE_SK, sbb_index=0, ipc_ip=IPC_IP, ipc_port=IPC_PORT)
        # Note we do not send a MakeNextBlockNotifcation

        # Before we SBB receives any tx batches, we expect his (one) manager to be empty
        self.assertEqual(len(sbb.sb_managers), 1)
        self.assertEqual(len(sbb.sb_managers[0].pending_txs), 0)

        # Send first TX bag in. This should get queued.
        tx_batch = SBBTester.create_tx_batch_env(num_txs=4, env_signing_key=MN_SK1)
        SBBTester.send_sub_to_sbb(sbb, envelope=tx_batch, handler_key=0)

        self.assertEqual(len(sbb.sb_managers[0].pending_txs), 1)
        self.assertEqual(sbb.sb_managers[0].pending_txs.popleft(), tx_batch)

    @SBBTester.test
    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_SUB_BLOCKS", 1)
    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_BLOCKS", 1)
    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_SB_BUILDERS", 1)
    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_SB_PER_BUILDER", 1)
    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_SB_PER_BLOCK_PER_BUILDER", 1)
    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.TRANSACTIONS_PER_SUB_BLOCK", 4)
    def test_double_tx_batch_builds_sbb(self, *args):
        sbb = SubBlockBuilder(ip=TEST_IP, signing_key=DELEGATE_SK, sbb_index=0, ipc_ip=IPC_IP, ipc_port=IPC_PORT)
        sbb._execute_next_sb = MagicMock()

        # First, fake a MakeNextBlock msg from BlockManager
        make_next_block = MakeNextBlock.create()
        SBBTester.send_ipc_to_sbb(sbb, make_next_block)

        # Send first TX bag in. Since w sent a MakeNextBlock notif, this should trigger a block construction
        tx_batch_env = SBBTester.create_tx_batch_env(num_txs=4, env_signing_key=MN_SK1)
        input_hash1 = Hasher.hash(tx_batch_env)
        SBBTester.send_sub_to_sbb(sbb, envelope=tx_batch_env, handler_key=0)

        # sbb._execute_next_sb.assert_called_with(input_hash1, tx_batch_env.message, 0)
        sbb._execute_next_sb.assert_called()


    @SBBTester.test
    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_SUB_BLOCKS", 1)
    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_BLOCKS", 1)
    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_SB_BUILDERS", 1)
    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_SB_PER_BUILDER", 1)
    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_SB_PER_BLOCK_PER_BUILDER", 1)
    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.TRANSACTIONS_PER_SUB_BLOCK", 4)
    def test_create_sbc_from_batch(self, *args):
        num_txs = 4
        input_hash = 'A' * 64
        sbb_idx = 0
        # create a few contract txs
        contract_txs, states, statuses = [], [], []
        for _ in range(num_txs):
            contract_txs.append(ContractTransactionBuilder.create_currency_tx('A' * 64, 'B' * 64, 100))
            states.append('SET MONEY OVER 9000')
            statuses.append('SUCC')

        sb_rep = [(c, status, state) for c, state, status, in zip(contract_txs, states, statuses)]

        sbb = SubBlockBuilder(ip=TEST_IP, signing_key=DELEGATE_SK, sbb_index=0, ipc_ip=IPC_IP, ipc_port=IPC_PORT)

        PhoneBook = VKBook(delegates=[sbb.verifying_key], masternodes=[TESTNET_MASTERNODES[0]['vk']])

        sbb._send_msg_over_ipc = MagicMock()

        exec_data = []
        for i in range(len(contract_txs)):
            exec_data.append(ExecutionData(contract=contract_txs[i], status=statuses[i],
                                           response='', state=states[i], stamps=10000))

        sb_data = SBData(input_hash=input_hash, tx_data=exec_data)

        sbb._create_sbc_from_batch(sb_data)

        # TODO assert tis was called with the expected SBC passed in
        sbb._send_msg_over_ipc.assert_called()

    @SBBTester.test
    # @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_SUB_BLOCKS", 1)
    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_SB_PER_BLOCK", 1)
    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_BLOCKS", 1)
    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_SB_BUILDERS", 1)
    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_SB_PER_BUILDER", 1)
    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_SB_PER_BLOCK_PER_BUILDER", 1)
    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.TRANSACTIONS_PER_SUB_BLOCK", 4)
    def test_execute_two_bags_one_builder(self, *args):
        sbb = SubBlockBuilder(ip=TEST_IP, signing_key=DELEGATE_SK, sbb_index=0, ipc_ip=IPC_IP, ipc_port=IPC_PORT)

        _, vk_1 = wallet.new(bytes.fromhex(MN_SK1))
        PhoneBook = VKBook(delegates=[sbb.verifying_key], masternodes=[vk_1])

        sbb._create_sbc_from_batch = MagicMock()

        # send intital make next blk notif
        make_next_block = MakeNextBlock.create()
        SBBTester.send_ipc_to_sbb(sbb, make_next_block)

        tx_batch_env = SBBTester.create_tx_batch_env(10, MN_SK1)
        SBBTester.send_sub_to_sbb(sbb, tx_batch_env, handler_key=0)

        # ANOTHER ONE
        tx_batch_env = SBBTester.create_tx_batch_env(10, MN_SK1)
        SBBTester.send_sub_to_sbb(sbb, tx_batch_env, handler_key=0)

        self.run_async(sbb.loop, 2)

        # now send him a make next blk
        make_next_block = MakeNextBlock.create()
        SBBTester.send_ipc_to_sbb(sbb, make_next_block)

        # make sure he called _created_sbc as a result of inputing a bag
        sbb._create_sbc_from_batch.assert_called_once()

        self.run_async(sbb.loop, 2)

        # yet ANOTHER ONE
        make_next_block = MakeNextBlock.create()
        SBBTester.send_ipc_to_sbb(sbb, make_next_block)

        self.run_async(sbb.loop, 2)
        self.assertEqual(sbb._create_sbc_from_batch.call_count, 2)

    # @SBBTester.test
    # # @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_SUB_BLOCKS", 2)
    # @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_SB_PER_BLOCK", 2)
    # @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_BLOCKS", 1)
    # @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_SB_BUILDERS", 2)
    # @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_SB_PER_BUILDER", 1)
    # @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_SB_PER_BLOCK_PER_BUILDER", 1)
    # @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.TRANSACTIONS_PER_SUB_BLOCK", 4)
    # def test_execute_two_bags_two_builder(self, *args):
    #     sbb1 = SubBlockBuilder(ip=TEST_IP, signing_key=DELEGATE_SK, sbb_index=0, ipc_ip=IPC_IP, ipc_port=IPC_PORT)
    #     sbb1._create_sbc_from_batch = MagicMock()
    #
    #     sbb2 = SubBlockBuilder(ip=TEST_IP, signing_key=DELEGATE_SK, sbb_index=0, ipc_ip=IPC_IP, ipc_port=IPC_PORT)
    #     sbb2._create_sbc_from_batch = MagicMock()
    #
    #     # send intital make next blk notif
    #     make_next_block = MakeNextBlock.create()
    #     SBBTester.send_ipc_to_sbb(sbb1, make_next_block)
    #     SBBTester.send_ipc_to_sbb(sbb2, make_next_block)
    #
    #     # THE FIRST
    #     tx_batch_env1 = SBBTester.create_tx_batch_env(10, MN_SK1)
    #     SBBTester.send_sub_to_sbb(sbb1, tx_batch_env1, handler_key=0)
    #     tx_batch_env2 = SBBTester.create_tx_batch_env(10, MN_SK2)
    #     SBBTester.send_sub_to_sbb(sbb2, tx_batch_env2, handler_key=0)
    #
    #     self.run_async(sbb1.loop, 2)
    #     self.run_async(sbb2.loop, 2)
    #
    #     # now send him a make next blk
    #     make_next_block = MakeNextBlock.create()
    #     SBBTester.send_ipc_to_sbb(sbb1, make_next_block)
    #     SBBTester.send_ipc_to_sbb(sbb2, make_next_block)
    #
    #     self.run_async(sbb1.loop, 2)
    #     self.run_async(sbb2.loop, 2)
    #
    #     # make sure he called _created_sbc as a result of inputing a bag
    #     sbb1._create_sbc_from_batch.assert_called_once()
    #     sbb2._create_sbc_from_batch.assert_called_once()
    #
    # @SBBTester.test
    # # @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_SUB_BLOCKS", 2)
    # @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_SB_PER_BLOCK", 2)
    # @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_BLOCKS", 1)
    # @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_SB_BUILDERS", 2)
    # @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_SB_PER_BUILDER", 1)
    # @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_SB_PER_BLOCK_PER_BUILDER", 1)
    # @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.TRANSACTIONS_PER_SUB_BLOCK", 4)
    # def test_execute_four_bags_two_builder(self, *args):
    #     sbb1 = SubBlockBuilder(ip=TEST_IP, signing_key=DELEGATE_SK, sbb_index=0, ipc_ip=IPC_IP, ipc_port=IPC_PORT)
    #     sbb1._create_sbc_from_batch = MagicMock()
    #
    #     sbb2 = SubBlockBuilder(ip=TEST_IP, signing_key=DELEGATE_SK, sbb_index=0, ipc_ip=IPC_IP, ipc_port=IPC_PORT)
    #     sbb2._create_sbc_from_batch = MagicMock()
    #
    #     # send intital make next blk notif
    #     make_next_block = MakeNextBlock.create()
    #     SBBTester.send_ipc_to_sbb(sbb1, make_next_block)
    #     SBBTester.send_ipc_to_sbb(sbb2, make_next_block)
    #
    #     tx_batch_env2 = SBBTester.create_tx_batch_env(20, MN_SK2)
    #     SBBTester.send_sub_to_sbb(sbb2, tx_batch_env2, handler_key=0)
    #
    #     # THE FIRST
    #     tx_batch_env1 = SBBTester.create_tx_batch_env(10, MN_SK1)
    #     SBBTester.send_sub_to_sbb(sbb1, tx_batch_env1, handler_key=0)
    #     tx_batch_env2 = SBBTester.create_tx_batch_env(20, MN_SK2)
    #     SBBTester.send_sub_to_sbb(sbb2, tx_batch_env2, handler_key=0)
    #
    #     self.run_async(sbb1.loop, 2)
    #     self.run_async(sbb2.loop, 2)
    #
    #     # now send him a make next blk
    #     make_next_block = MakeNextBlock.create()
    #     SBBTester.send_ipc_to_sbb(sbb1, make_next_block)
    #     SBBTester.send_ipc_to_sbb(sbb2, make_next_block)
    #
    #     self.run_async(sbb1.loop, 2)
    #     self.run_async(sbb2.loop, 2)
    #
    #     # make sure he called _created_sbc as a result of inputing a bag
    #     sbb1._create_sbc_from_batch.assert_called_once()
    #     sbb2._create_sbc_from_batch.assert_called_once()
    #
    #     self.run_async(sbb1.loop, 2)
    #     self.run_async(sbb2.loop, 2)
    #
    #     # # yet ANOTHER ONE
    #     # make_next_block = MakeNextBlock.create()
    #     # SBBTester.send_ipc_to_sbb(sbb1, make_next_block)
    #     # SBBTester.send_ipc_to_sbb(sbb2, make_next_block)
    #
    #     # THE SECOND
    #     tx_batch_env1 = SBBTester.create_tx_batch_env(10, MN_SK1)
    #     SBBTester.send_sub_to_sbb(sbb1, tx_batch_env1, handler_key=0)
    #
    #
    #     self.run_async(sbb1.loop, 2)
    #     self.run_async(sbb2.loop, 2)
    #     self.assertEqual(sbb1._create_sbc_from_batch.call_count, 2)
    #     self.assertEqual(sbb2._create_sbc_from_batch.call_count, 2)
    #
    # @SBBTester.test
    # # @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_SUB_BLOCKS", 2)
    # @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_SB_PER_BLOCK", 2)
    # @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_BLOCKS", 1)
    # @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_SB_BUILDERS", 2)
    # @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_SB_PER_BUILDER", 1)
    # @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_SB_PER_BLOCK_PER_BUILDER", 1)
    # @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.TRANSACTIONS_PER_SUB_BLOCK", 4)
    # def test_execute_hella_bags_two_builder(self, *args):
    #     num_bags = 2
    #     bags_for_1 = []
    #     bags_for_2 = []
    #
    #     for _ in range(num_bags):
    #         tx_batch_env1 = SBBTester.create_tx_batch_env(10, MN_SK1)
    #         tx_batch_env2 = SBBTester.create_tx_batch_env(10, MN_SK1)
    #         bags_for_1.append(tx_batch_env1)
    #         bags_for_2.append(tx_batch_env2)
    #
    #     sbb1 = SubBlockBuilder(ip=TEST_IP, signing_key=DELEGATE_SK, sbb_index=0, ipc_ip=IPC_IP, ipc_port=IPC_PORT)
    #     sbb1._create_sbc_from_batch = MagicMock()
    #
    #     sbb2 = SubBlockBuilder(ip=TEST_IP, signing_key=DELEGATE_SK, sbb_index=0, ipc_ip=IPC_IP, ipc_port=IPC_PORT)
    #     sbb2._create_sbc_from_batch = MagicMock()
    #
    #     # send intital make next blk notif
    #     make_next_block = MakeNextBlock.create()
    #     SBBTester.send_ipc_to_sbb(sbb1, make_next_block)
    #     SBBTester.send_ipc_to_sbb(sbb2, make_next_block)
    #
    #     # dump all bags on them
    #     for i in range(num_bags):
    #         SBBTester.send_sub_to_sbb(sbb1, bags_for_1[i], handler_key=0)
    #         SBBTester.send_sub_to_sbb(sbb2, bags_for_2[i], handler_key=0)
    #         self.run_async(sbb1.loop, 0.5)
    #         self.run_async(sbb2.loop, 0.5)
    #
    #     self.run_async(sbb1.loop, 1)
    #     self.run_async(sbb2.loop, 1)
    #
    #     # give them the merge calls
    #     for i in range(num_bags):
    #         make_next_block = MakeNextBlock.create()
    #         SBBTester.send_ipc_to_sbb(sbb1, make_next_block)
    #         SBBTester.send_ipc_to_sbb(sbb2, make_next_block)
    #         self.run_async(sbb1.loop, 0.5)
    #         self.run_async(sbb2.loop, 0.5)
    #
    #     self.run_async(sbb1.loop, 1)
    #     self.run_async(sbb2.loop, 1)
    #
    #     self.assertEqual(sbb1._create_sbc_from_batch.call_count, num_bags)
    #     self.assertEqual(sbb2._create_sbc_from_batch.call_count, num_bags)

    @SBBTester.test
    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_SUB_BLOCKS", 1)
    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_BLOCKS", 1)
    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_SB_BUILDERS", 1)
    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_SB_PER_BUILDER", 1)
    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_SB_PER_BLOCK", 1)
    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.NUM_SB_PER_BLOCK_PER_BUILDER", 1)
    @mock.patch("cilantro_ee.nodes.delegate.sub_block_builder.TRANSACTIONS_PER_SUB_BLOCK", 4)
    def test_blk_notification_handler(self, *args):
        sbb = SubBlockBuilder(ip=TEST_IP, signing_key=DELEGATE_SK, sbb_index=0, ipc_ip=IPC_IP, ipc_port=IPC_PORT)
        self.assertEqual(len(sbb.sb_managers), 1)
        sbb._send_msg_over_ipc = MagicMock()
        sbb._create_empty_sbc = MagicMock()
        sbb._create_sbc_from_batch = MagicMock()

        # send it 6 blocks w/ 4 txns each except second block (index = 1) is empty
        ib_hashes = []
        for i in range(6):
            num_txs = 0 if i == 1 else 4
            tx_batch = SBBTester.create_tx_batch_env(num_txs=num_txs, env_signing_key=MN_SK1)
            input_hash = Hasher.hash(tx_batch)
            ib_hashes.append(input_hash)
            print("{} {}".format(i, input_hash))
            SBBTester.send_sub_to_sbb(sbb, envelope=tx_batch, handler_key=0)

        # Now, send a MakeNextBlock notif. This should trigger an empty subblock to be built and sent over IPC to BM
        make_next_block = MakeNextBlock.create()
        SBBTester.send_ipc_to_sbb(sbb, make_next_block)

        self.run_async(sbb.loop, 2)
        # should receive a _new subblock for input 0 and assert its input_hash is same as the one above
        # SBBTester.send_ipc_to_sbb(sbb, make_next_block)
        sbb._create_sbc_from_batch.assert_called_once()
        sb = sbb._create_sbc_from_batch.call_args[0][0]
        self.assertEqual(sb.input_hash, ib_hashes[0])

        SBBTester.send_ipc_to_sbb(sbb, make_next_block)
        self.run_async(sbb.loop, 2)
        # should receive a _new subblock for input 1
        sbb._create_empty_sbc.assert_called()
        sb = sbb._create_empty_sbc.call_args[0][0]
        self.assertEqual(sb.input_hash, ib_hashes[1])

        # send align notification for input hash # 3
        # message = AlignInputHash.create(ib_hashes[3], 0)
        # message = AlignInputHash.create(ib_hashes[3], sb_index=0)
        # SBBTester.send_ipc_to_sbb(sbb, message)
        # SBBTester.send_ipc_to_sbb(sbb, make_next_block)
        # self.run_async(sbb.loop, 2)
        # should receive a _new subblock for input 4
        # sbb._create_sbc_from_batch.assert_called()
        # sb = sbb._create_sbc_from_batch.call_args[0][0]
        # self.assertEqual(sb.input_hash, ib_hashes[4])

        # send failedblk notification w/ input hashes 3, 4 
        fih = []
        fih.append(ib_hashes[3])
        fih.append(ib_hashes[4])
        ffih = []
        ffih.append(fih)
        message = FailedBlockNotification.create(ib_hashes[4], ffih)
        SBBTester.send_ipc_to_sbb(sbb, message)
        SBBTester.send_ipc_to_sbb(sbb, make_next_block)
        self.run_async(sbb.loop, 2)
        # should receive a _new subblock for input 5
        sbb._create_sbc_from_batch.assert_called()
        sb = sbb._create_sbc_from_batch.call_args[0][0]
        self.assertEqual(sb.input_hash, ib_hashes[5])


if __name__ == "__main__":
    import unittest
    unittest.main()



from unittest import TestCase
from cilantro_ee.core.nonces import NonceManager
from cilantro_ee.core.utils.transaction import TransactionBuilder, transaction_is_valid
from cilantro_ee.core.utils import transaction
from cilantro_ee.core.crypto.wallet import Wallet
from cilantro_ee.core.messages.capnp_impl import capnp_struct as schemas
from contracting import config
import secrets
import os
import capnp

transaction_capnp = capnp.load(os.path.dirname(schemas.__file__) + '/transaction.capnp')


class TestTXValidity(TestCase):
    def setUp(self):
        self.nonce_manager = NonceManager()
        self.nonce_manager.driver.flush()

    def tearDown(self):
        self.nonce_manager.driver.flush()

    def test_processor_incorrect_returns_false(self):
        w = Wallet()
        expected_processor = secrets.token_bytes(32)
        given_processor = secrets.token_bytes(32)

        tx = TransactionBuilder(w.verifying_key(),
                                contract='currency',
                                function='transfer',
                                kwargs={'amount': 10, 'to': 'jeff'},
                                stamps=500000,
                                processor=given_processor,
                                nonce=0)

        tx.sign(w.signing_key())
        tx_bytes = tx.serialize()
        tx_struct = transaction_capnp.Transaction.from_bytes_packed(tx_bytes)

        with self.assertRaises(transaction.TransactionProcessorInvalid):
            transaction_is_valid(tx=tx_struct, expected_processor=expected_processor, driver=self.nonce_manager)

    def test_processor_is_expected_but_nonce_is_incorrect_returns_false(self):
        w = Wallet()
        expected_processor = secrets.token_bytes(32)

        tx = TransactionBuilder(w.verifying_key(),
                                contract='currency',
                                function='transfer',
                                kwargs={'amount': 10, 'to': 'jeff'},
                                stamps=500000,
                                processor=expected_processor,
                                nonce=1)

        tx.sign(w.signing_key())
        tx_bytes = tx.serialize()
        tx_struct = transaction_capnp.Transaction.from_bytes_packed(tx_bytes)

        with self.assertRaises(transaction.TransactionNonceInvalid):
            transaction_is_valid(tx=tx_struct, expected_processor=expected_processor, driver=self.nonce_manager)

    def test_processor_and_nonce_correct_increments_pending_nonce_by_one(self):
        w = Wallet()
        expected_processor = secrets.token_bytes(32)

        tx = TransactionBuilder(w.verifying_key(),
                                contract='currency',
                                function='transfer',
                                kwargs={'amount': 10, 'to': 'jeff'},
                                stamps=0,
                                processor=expected_processor,
                                nonce=0)

        tx.sign(w.signing_key())
        tx_bytes = tx.serialize()
        tx_struct = transaction_capnp.Transaction.from_bytes_packed(tx_bytes)

        transaction_is_valid(tx=tx_struct, expected_processor=expected_processor, driver=self.nonce_manager)

        pending_nonce = self.nonce_manager.get_pending_nonce(expected_processor, w.verifying_key())

        self.assertEqual(pending_nonce, 1)

    def test_all_but_wallet_signed_returns_false(self):
        w = Wallet()
        x = Wallet()

        expected_processor = secrets.token_bytes(32)

        tx = TransactionBuilder(w.verifying_key(),
                                contract='currency',
                                function='transfer',
                                kwargs={'amount': 10, 'to': 'jeff'},
                                stamps=0,
                                processor=expected_processor,
                                nonce=0)

        tx.sign(x.signing_key())
        tx_bytes = tx.serialize()
        tx_struct = transaction_capnp.Transaction.from_bytes_packed(tx_bytes)

        with self.assertRaises(transaction.TransactionSignatureInvalid):
            transaction_is_valid(tx=tx_struct, expected_processor=expected_processor, driver=self.nonce_manager)

    def test_all_but_proof_valid_returns_false(self):
        w = Wallet()

        expected_processor = secrets.token_bytes(32)

        tx = TransactionBuilder(w.verifying_key(),
                                contract='currency',
                                function='transfer',
                                kwargs={'amount': 10, 'to': 'jeff'},
                                stamps=0,
                                processor=expected_processor,
                                nonce=0)

        tx.proof = b'\00' * 32
        tx.proof_generated = True

        tx.sign(w.signing_key())
        tx_bytes = tx.serialize()
        tx_struct = transaction_capnp.Transaction.from_bytes_packed(tx_bytes)

        with self.assertRaises(transaction.TransactionPOWProofInvalid):
            transaction_is_valid(tx=tx_struct, expected_processor=expected_processor, driver=self.nonce_manager)

    def test_processor_and_nonce_correct_but_not_enough_stamps_returns_false(self):
        w = Wallet()
        expected_processor = secrets.token_bytes(32)

        tx = TransactionBuilder(w.verifying_key(),
                                contract='currency',
                                function='transfer',
                                kwargs={'amount': 10, 'to': 'jeff'},
                                stamps=500000,
                                processor=expected_processor,
                                nonce=0)

        tx.sign(w.signing_key())
        tx_bytes = tx.serialize()
        tx_struct = transaction_capnp.Transaction.from_bytes_packed(tx_bytes)

        with self.assertRaises(transaction.TransactionSenderTooFewStamps):
            transaction_is_valid(tx=tx_struct, expected_processor=expected_processor, driver=self.nonce_manager)

        balances_key = '{}{}{}{}{}'.format('currency',
                                           config.INDEX_SEPARATOR,
                                           'balances',
                                           config.DELIMITER,
                                           tx.payload.sender.hex())

        balance = self.nonce_manager.driver.get(balances_key) or 0

        self.assertEqual(balance, 0)

    def test_all_valid_with_stamps_when_balance_is_set(self):
        w = Wallet()
        expected_processor = secrets.token_bytes(32)

        tx = TransactionBuilder(w.verifying_key(),
                                contract='currency',
                                function='transfer',
                                kwargs={'amount': 10, 'to': 'jeff'},
                                stamps=500000,
                                processor=expected_processor,
                                nonce=0)

        tx.sign(w.signing_key())
        tx_bytes = tx.serialize()
        tx_struct = transaction_capnp.Transaction.from_bytes_packed(tx_bytes)

        balances_key = '{}{}{}{}{}'.format('currency',
                                           config.INDEX_SEPARATOR,
                                           'balances',
                                           config.DELIMITER,
                                           tx.payload.sender.hex())

        self.nonce_manager.driver.set(balances_key, 500000)

        transaction_is_valid(tx=tx_struct, expected_processor=expected_processor, driver=self.nonce_manager)
        balance = self.nonce_manager.driver.get(balances_key) or 0

        self.assertEqual(balance, 500000)

    def test_multiple_nonces_in_sequence_all_verify(self):
        w = Wallet()
        expected_processor = secrets.token_bytes(32)

        balances_key = '{}{}{}{}{}'.format('currency',
                                           config.INDEX_SEPARATOR,
                                           'balances',
                                           config.DELIMITER,
                                           w.verifying_key().hex())

        self.nonce_manager.driver.set(balances_key, 500000)

        tx = TransactionBuilder(w.verifying_key(),
                                contract='currency',
                                function='transfer',
                                kwargs={'amount': 10, 'to': 'jeff'},
                                stamps=500000,
                                processor=expected_processor,
                                nonce=0)

        tx.sign(w.signing_key())
        tx_bytes = tx.serialize()
        tx_struct = transaction_capnp.Transaction.from_bytes_packed(tx_bytes)

        transaction_is_valid(tx=tx_struct, expected_processor=expected_processor, driver=self.nonce_manager)

        tx = TransactionBuilder(w.verifying_key(),
                                contract='currency',
                                function='transfer',
                                kwargs={'amount': 10, 'to': 'jeff'},
                                stamps=500000,
                                processor=expected_processor,
                                nonce=1)

        tx.sign(w.signing_key())
        tx_bytes = tx.serialize()
        tx_struct = transaction_capnp.Transaction.from_bytes_packed(tx_bytes)

        transaction_is_valid(tx=tx_struct, expected_processor=expected_processor, driver=self.nonce_manager)

        tx = TransactionBuilder(w.verifying_key(),
                                contract='currency',
                                function='transfer',
                                kwargs={'amount': 10, 'to': 'jeff'},
                                stamps=500000,
                                processor=expected_processor,
                                nonce=2)

        tx.sign(w.signing_key())
        tx_bytes = tx.serialize()
        tx_struct = transaction_capnp.Transaction.from_bytes_packed(tx_bytes)

        transaction_is_valid(tx=tx_struct, expected_processor=expected_processor, driver=self.nonce_manager)

        tx = TransactionBuilder(w.verifying_key(),
                                contract='currency',
                                function='transfer',
                                kwargs={'amount': 10, 'to': 'jeff'},
                                stamps=500000,
                                processor=expected_processor,
                                nonce=3)

        tx.sign(w.signing_key())
        tx_bytes = tx.serialize()
        tx_struct = transaction_capnp.Transaction.from_bytes_packed(tx_bytes)

        transaction_is_valid(tx=tx_struct, expected_processor=expected_processor, driver=self.nonce_manager)

    def test_non_sequence_fails_in_strict_mode(self):
        w = Wallet()
        expected_processor = secrets.token_bytes(32)

        balances_key = '{}{}{}{}{}'.format('currency',
                                           config.INDEX_SEPARATOR,
                                           'balances',
                                           config.DELIMITER,
                                           w.verifying_key().hex())

        self.nonce_manager.driver.set(balances_key, 500000)

        tx = TransactionBuilder(w.verifying_key(),
                                contract='currency',
                                function='transfer',
                                kwargs={'amount': 10, 'to': 'jeff'},
                                stamps=500000,
                                processor=expected_processor,
                                nonce=0)

        tx.sign(w.signing_key())
        tx_bytes = tx.serialize()
        tx_struct = transaction_capnp.Transaction.from_bytes_packed(tx_bytes)

        transaction_is_valid(tx=tx_struct, expected_processor=expected_processor, driver=self.nonce_manager)

        tx = TransactionBuilder(w.verifying_key(),
                                contract='currency',
                                function='transfer',
                                kwargs={'amount': 10, 'to': 'jeff'},
                                stamps=500000,
                                processor=expected_processor,
                                nonce=1)

        tx.sign(w.signing_key())
        tx_bytes = tx.serialize()
        tx_struct = transaction_capnp.Transaction.from_bytes_packed(tx_bytes)

        transaction_is_valid(tx=tx_struct, expected_processor=expected_processor, driver=self.nonce_manager)

        tx = TransactionBuilder(w.verifying_key(),
                                contract='currency',
                                function='transfer',
                                kwargs={'amount': 10, 'to': 'jeff'},
                                stamps=500000,
                                processor=expected_processor,
                                nonce=2)

        tx.sign(w.signing_key())
        tx_bytes = tx.serialize()
        tx_struct = transaction_capnp.Transaction.from_bytes_packed(tx_bytes)

        transaction_is_valid(tx=tx_struct, expected_processor=expected_processor, driver=self.nonce_manager)

        tx = TransactionBuilder(w.verifying_key(),
                                contract='currency',
                                function='transfer',
                                kwargs={'amount': 10, 'to': 'jeff'},
                                stamps=500000,
                                processor=expected_processor,
                                nonce=5)

        tx.sign(w.signing_key())
        tx_bytes = tx.serialize()
        tx_struct = transaction_capnp.Transaction.from_bytes_packed(tx_bytes)

        with self.assertRaises(transaction.TransactionNonceInvalid):
            transaction_is_valid(tx=tx_struct, expected_processor=expected_processor, driver=self.nonce_manager)

    def test_greater_than_passes_in_no_strict_mode(self):
        w = Wallet()
        expected_processor = secrets.token_bytes(32)

        balances_key = '{}{}{}{}{}'.format('currency',
                                           config.INDEX_SEPARATOR,
                                           'balances',
                                           config.DELIMITER,
                                           w.verifying_key().hex())

        self.nonce_manager.driver.set(balances_key, 500000)

        tx = TransactionBuilder(w.verifying_key(),
                                contract='currency',
                                function='transfer',
                                kwargs={'amount': 10, 'to': 'jeff'},
                                stamps=500000,
                                processor=expected_processor,
                                nonce=0)

        tx.sign(w.signing_key())
        tx_bytes = tx.serialize()
        tx_struct = transaction_capnp.Transaction.from_bytes_packed(tx_bytes)

        transaction_is_valid(tx=tx_struct, expected_processor=expected_processor, driver=self.nonce_manager)

        tx = TransactionBuilder(w.verifying_key(),
                                contract='currency',
                                function='transfer',
                                kwargs={'amount': 10, 'to': 'jeff'},
                                stamps=500000,
                                processor=expected_processor,
                                nonce=1)

        tx.sign(w.signing_key())
        tx_bytes = tx.serialize()
        tx_struct = transaction_capnp.Transaction.from_bytes_packed(tx_bytes)

        transaction_is_valid(tx=tx_struct, expected_processor=expected_processor, driver=self.nonce_manager)

        tx = TransactionBuilder(w.verifying_key(),
                                contract='currency',
                                function='transfer',
                                kwargs={'amount': 10, 'to': 'jeff'},
                                stamps=500000,
                                processor=expected_processor,
                                nonce=2)

        tx.sign(w.signing_key())
        tx_bytes = tx.serialize()
        tx_struct = transaction_capnp.Transaction.from_bytes_packed(tx_bytes)

        transaction_is_valid(tx=tx_struct, expected_processor=expected_processor, driver=self.nonce_manager)

        tx = TransactionBuilder(w.verifying_key(),
                                contract='currency',
                                function='transfer',
                                kwargs={'amount': 10, 'to': 'jeff'},
                                stamps=500000,
                                processor=expected_processor,
                                nonce=5)

        tx.sign(w.signing_key())
        tx_bytes = tx.serialize()
        tx_struct = transaction_capnp.Transaction.from_bytes_packed(tx_bytes)

        transaction_is_valid(tx=tx_struct, expected_processor=expected_processor, driver=self.nonce_manager,
                             strict=False)

    def test_non_strict_fails_if_same_nonce(self):
        w = Wallet()
        expected_processor = secrets.token_bytes(32)

        balances_key = '{}{}{}{}{}'.format('currency',
                                           config.INDEX_SEPARATOR,
                                           'balances',
                                           config.DELIMITER,
                                           w.verifying_key().hex())

        self.nonce_manager.driver.set(balances_key, 500000)

        tx = TransactionBuilder(w.verifying_key(),
                                contract='currency',
                                function='transfer',
                                kwargs={'amount': 10, 'to': 'jeff'},
                                stamps=500000,
                                processor=expected_processor,
                                nonce=0)

        tx.sign(w.signing_key())
        tx_bytes = tx.serialize()
        tx_struct = transaction_capnp.Transaction.from_bytes_packed(tx_bytes)

        transaction_is_valid(tx=tx_struct, expected_processor=expected_processor, driver=self.nonce_manager,
                             strict=False)

        tx = TransactionBuilder(w.verifying_key(),
                                contract='currency',
                                function='transfer',
                                kwargs={'amount': 10, 'to': 'jeff'},
                                stamps=500000,
                                processor=expected_processor,
                                nonce=0)

        tx.sign(w.signing_key())
        tx_bytes = tx.serialize()
        tx_struct = transaction_capnp.Transaction.from_bytes_packed(tx_bytes)

        with self.assertRaises(transaction.TransactionNonceInvalid):
            transaction_is_valid(tx=tx_struct, expected_processor=expected_processor, driver=self.nonce_manager,
                                 strict=False)

    def test_strict_fails_if_same_nonce(self):
        w = Wallet()
        expected_processor = secrets.token_bytes(32)

        balances_key = '{}{}{}{}{}'.format('currency',
                                           config.INDEX_SEPARATOR,
                                           'balances',
                                           config.DELIMITER,
                                           w.verifying_key().hex())

        self.nonce_manager.driver.set(balances_key, 500000)

        tx = TransactionBuilder(w.verifying_key(),
                                contract='currency',
                                function='transfer',
                                kwargs={'amount': 10, 'to': 'jeff'},
                                stamps=500000,
                                processor=expected_processor,
                                nonce=0)

        tx.sign(w.signing_key())
        tx_bytes = tx.serialize()
        tx_struct = transaction_capnp.Transaction.from_bytes_packed(tx_bytes)

        transaction_is_valid(tx=tx_struct, expected_processor=expected_processor, driver=self.nonce_manager)

        tx = TransactionBuilder(w.verifying_key(),
                                contract='currency',
                                function='transfer',
                                kwargs={'amount': 10, 'to': 'jeff'},
                                stamps=500000,
                                processor=expected_processor,
                                nonce=0)

        tx.sign(w.signing_key())
        tx_bytes = tx.serialize()
        tx_struct = transaction_capnp.Transaction.from_bytes_packed(tx_bytes)

        with self.assertRaises(transaction.TransactionNonceInvalid):
            transaction_is_valid(tx=tx_struct, expected_processor=expected_processor, driver=self.nonce_manager)

    def test_tx_nonce_minus_nonce_less_than_tx_per_block(self):
        w = Wallet()
        expected_processor = secrets.token_bytes(32)

        balances_key = '{}{}{}{}{}'.format('currency',
                                           config.INDEX_SEPARATOR,
                                           'balances',
                                           config.DELIMITER,
                                           w.verifying_key().hex())

        self.nonce_manager.driver.set(balances_key, 500000)

        self.nonce_manager.set_nonce(processor=expected_processor, sender=w.verifying_key(), nonce=1)
        self.nonce_manager.set_pending_nonce(processor=expected_processor, sender=w.verifying_key(), nonce=2)

        tx = TransactionBuilder(w.verifying_key(),
                                contract='currency',
                                function='transfer',
                                kwargs={'amount': 10, 'to': 'jeff'},
                                stamps=500000,
                                processor=expected_processor,
                                nonce=2)

        tx.sign(w.signing_key())
        tx_bytes = tx.serialize()
        tx_struct = transaction_capnp.Transaction.from_bytes_packed(tx_bytes)

        transaction_is_valid(tx=tx_struct, expected_processor=expected_processor, driver=self.nonce_manager)

    def test_tx_nonce_minus_nonce_greater_than_tx_per_block_fails(self):
        w = Wallet()
        expected_processor = secrets.token_bytes(32)

        balances_key = '{}{}{}{}{}'.format('currency',
                                           config.INDEX_SEPARATOR,
                                           'balances',
                                           config.DELIMITER,
                                           w.verifying_key().hex())

        self.nonce_manager.driver.set(balances_key, 500000)

        self.nonce_manager.set_nonce(processor=expected_processor, sender=w.verifying_key(), nonce=1)
        self.nonce_manager.set_pending_nonce(processor=expected_processor, sender=w.verifying_key(), nonce=10)

        tx = TransactionBuilder(w.verifying_key(),
                                contract='currency',
                                function='transfer',
                                kwargs={'amount': 10, 'to': 'jeff'},
                                stamps=500000,
                                processor=expected_processor,
                                nonce=20)

        tx.sign(w.signing_key())
        tx_bytes = tx.serialize()
        tx_struct = transaction_capnp.Transaction.from_bytes_packed(tx_bytes)

        with self.assertRaises(transaction.TransactionTooManyPendingException):
            transaction_is_valid(tx=tx_struct, expected_processor=expected_processor, driver=self.nonce_manager)

    def test_nonce_minus_pending_nonce_equal_tx_per_block_fails(self):
        w = Wallet()
        expected_processor = secrets.token_bytes(32)

        balances_key = '{}{}{}{}{}'.format('currency',
                                           config.INDEX_SEPARATOR,
                                           'balances',
                                           config.DELIMITER,
                                           w.verifying_key().hex())

        self.nonce_manager.driver.set(balances_key, 500000)

        self.nonce_manager.set_nonce(processor=expected_processor, sender=w.verifying_key(), nonce=1)
        self.nonce_manager.set_pending_nonce(processor=expected_processor, sender=w.verifying_key(), nonce=16)

        tx = TransactionBuilder(w.verifying_key(),
                                contract='currency',
                                function='transfer',
                                kwargs={'amount': 10, 'to': 'jeff'},
                                stamps=500000,
                                processor=expected_processor,
                                nonce=16)

        tx.sign(w.signing_key())
        tx_bytes = tx.serialize()
        tx_struct = transaction_capnp.Transaction.from_bytes_packed(tx_bytes)

        with self.assertRaises(transaction.TransactionTooManyPendingException):
            transaction_is_valid(tx=tx_struct, expected_processor=expected_processor, driver=self.nonce_manager)

    def test_15_in_row_valid_16th_not_due_to_tx_per_block_failing(self):
        w = Wallet()
        expected_processor = secrets.token_bytes(32)

        balances_key = '{}{}{}{}{}'.format('currency',
                                           config.INDEX_SEPARATOR,
                                           'balances',
                                           config.DELIMITER,
                                           w.verifying_key().hex())

        self.nonce_manager.driver.set(balances_key, 500000)

        for i in range(15):
            tx = TransactionBuilder(w.verifying_key(),
                                    contract='currency',
                                    function='transfer',
                                    kwargs={'amount': 10, 'to': 'jeff'},
                                    stamps=500000,
                                    processor=expected_processor,
                                    nonce=i)

            tx.sign(w.signing_key())
            tx_bytes = tx.serialize()
            tx_struct = transaction_capnp.Transaction.from_bytes_packed(tx_bytes)

            transaction_is_valid(tx=tx_struct, expected_processor=expected_processor, driver=self.nonce_manager)

        tx = TransactionBuilder(w.verifying_key(),
                                contract='currency',
                                function='transfer',
                                kwargs={'amount': 10, 'to': 'jeff'},
                                stamps=500000,
                                processor=expected_processor,
                                nonce=15)

        tx.sign(w.signing_key())
        tx_bytes = tx.serialize()
        tx_struct = transaction_capnp.Transaction.from_bytes_packed(tx_bytes)

        with self.assertRaises(transaction.TransactionTooManyPendingException):
            transaction_is_valid(tx=tx_struct, expected_processor=expected_processor, driver=self.nonce_manager)



"""
    BlockManager  (main process of delegate)

    This is the main workhorse for managing inter node communication as well as
    coordinating the interpreting and creation of sub-block contenders that form part of new block.
    It creates sub-block builder processes to manage the parallel execution of different sub-blocks.
    It will also participate in conflict resolution of sub-blocks
    It publishes those sub-blocks to masters so they can assemble the new block contender

    It manages the new block notifications from master and update the db snapshot state
    so sub-block builders can proceed to next block

"""

from cilantro_ee.core.logger.base import get_logger

from cilantro_ee.nodes.catchup import CatchupManager
from cilantro_ee.nodes.delegate.sub_block_builder import SubBlockBuilder

from cilantro_ee.services.storage.state import MetaDataStorage

from cilantro_ee.core.utils.worker import Worker

from cilantro_ee.utils.lprocess import LProcess

from cilantro_ee.constants.system_config import *
from cilantro_ee.constants.zmq_filters import DEFAULT_FILTER, NEW_BLK_NOTIF_FILTER
from cilantro_ee.constants.ports import *
from cilantro_ee.constants import conf

from cilantro_ee.core.messages.message_type import MessageType
from cilantro_ee.core.messages.message import Message

from cilantro_ee.core.crypto.wallet import _verify
from cilantro_ee.contracts.sync import sync_genesis_contracts
import hashlib
import asyncio, zmq, time, random
import os

IPC_IP = 'block-manager-ipc-sock'
IPC_PORT = 6967

from cilantro_ee.core.nonces import NonceManager


# class to keep track of sub-blocks sent over from my sub-block builders
class SubBlocks:
    def __init__(self):
        self.sbs = {}
        self.futures = []

    def reset(self):
        self.sbs = {}
        self.futures = []

    def is_quorum(self):
        return len(self.sbs) == NUM_SB_PER_BLOCK

    def add_sub_block(self, sub_block, fut):
        sb_idx = sub_block.subBlockIdx % NUM_SB_PER_BLOCK

        if sb_idx in self.sbs:
            # todo log it as an issue
            pass

        self.sbs[sb_idx] = sub_block

        self.futures.append(fut)

    def get_sb_hashes_sorted(self):
        sb_hashes = []

        for i in range(NUM_SB_PER_BLOCK):
            sb = self.sbs[i]
            sb_hashes.append(sb.resultHash)

        return sb_hashes

    def get_input_hashes_sorted(self):
        sb_hashes = []
        num_sbs = len(self.sbs)

        for i in range(num_sbs):
            sb = self.sbs[i]
            sb_hashes.append(sb.inputHash.hex())

        return sb_hashes


class NextBlockData:
    def __init__(self, block_notif):
        self.block_notif = block_notif
        is_failed = block_notif.which() == "FailedBlock"
        self.quorum_num = FAILED_BLOCK_NOTIFICATION_QUORUM if is_failed \
                            else BLOCK_NOTIFICATION_QUORUM
        self.is_quorum = False
        self.senders = set()

    # TODO: Deprecate. Not used in project files
    def is_quorum(self):
        return self.is_quorum

    def add_sender(self, sender):
        self.senders.add(sender)

        if not self.is_quorum and (len(self.senders) >= self.quorum_num):
            self.is_quorum = True
            return True

        return False


# Keeps track of block notifications from master
class NextBlock:
    def __init__(self):
        self.next_block_data = {}
        self.quorum_block = None
        self.hard_reset()

    # use this when it has to go to catchup
    def hard_reset(self):
        self.next_block_data = {}     # hash of block num -> block hash -> data
        self.quorum_block = None

    def reset(self, block_num):
        if self.quorum_block:
            bn = self.quorum_block.blockNum
            if bn < block_num and bn in self.next_block_data:
                try:
                    del self.next_block_data[bn] # This can never happen because you are checking for the key prior to deleting
                except KeyError:
                    pass
                    # todo add a debug message - not supposed to happen

        self.quorum_block = None

    def is_quorum(self):
        return self.quorum_block != None

    def get_quorum_block(self):
        return self.quorum_block

    def add_notification(self, block_notif, sender, block_num, block_hash):
        if self.quorum_block and (self.quorum_block.blockNum == block_num):
            # todo - if it is not matching blockhash, may need to audit it
            return False

        if block_num not in self.next_block_data: # default dict?
            self.next_block_data[block_num] = {}
            # todo add time info to implement timeout

        if block_hash not in self.next_block_data[block_num]:
            self.next_block_data[block_num][block_hash] = NextBlockData(block_notif)

        if self.next_block_data[block_num][block_hash].add_sender(sender):
            self.quorum_block = block_notif
            return True

        return False


class DBState:
    CATCHUP = 'in_catchup_phase'
    CURRENT = 'up_to_date'

    """ convenience struct to maintain db snapshot state data in one place """
    def __init__(self):
        self.driver = MetaDataStorage()
        self.next_block = NextBlock()
        self.my_sub_blocks = SubBlocks()

        self.catchup_mgr = None
        self.is_catchup_done = False

    def reset(self):
        # reset all the state info
        self.next_block.reset(self.driver.latest_block_num)
        self.my_sub_blocks.reset()


class BlockManager(Worker):
    def __init__(self, ip, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.log = get_logger("BlockManager[{}]".format(self.verifying_key[:8]))

        self.ip = ip
        self.sb_builders = {}  # index -> process
        # raghu todo - delete this and remove sb_index related functionality
        self.sb_index = PhoneBook.delegates.index(self.wallet.verifying_key().hex()) % NUM_SB_BUILDERS

        self.sbb_not_ready_count = NUM_SB_BUILDERS

        self.db_state = DBState()
        self.my_quorum = PhoneBook.masternode_quorum_min
        self._pending_work_at_sbb = 0
        self._masternodes_ready = set()
        self.start_sub_blocks = 0

        self._thicc_log()

        # Define Sockets (these get set in build_task_list)
        self.router, self.ipc_router, self.pub, self.sub, self.nbn_sub = None, None, None, None, None

        self.ipc_ip = IPC_IP + '-' + str(os.getpid()) + '-' + str(random.randint(0, 2**32))

        self.driver = MetaDataStorage()
        self.nonce_manager = NonceManager()
        self.run()

    def _thicc_log(self):
        self.log.notice("\nBlockManager initializing with\nvk={vk}\nsubblock_index={sb_index}\n"
                        "num_sub_blocks={num_sb}\nnum_blocks={num_blocks}\nsub_blocks_per_block={sb_per_block}\n"
                        "num_sb_builders={num_sb_builders}\nsub_blocks_per_builder={sb_per_builder}\n"
                        "sub_blocks_per_block_per_builder={sb_per_block_per_builder}\n"
                        .format(vk=self.verifying_key, sb_index=self.sb_index, num_sb=NUM_SUB_BLOCKS,
                                num_blocks=NUM_BLOCKS, sb_per_block=NUM_SB_PER_BLOCK,
                                num_sb_builders=NUM_SB_BUILDERS, sb_per_builder=NUM_SB_PER_BUILDER,
                                sb_per_block_per_builder=NUM_SB_PER_BLOCK_PER_BUILDER))

    def run(self):
        self.build_task_list()
        self.log.info("Block Manager starting...")
        self.start_sbb_procs()

        self.loop.run_until_complete(asyncio.gather(*self.tasks))

    def set_pending_work(self, sbb_index):
        # self.log.info("set pending work {} for sbb {}".format(self._pending_work_at_sbb, sbb_index))
        self._pending_work_at_sbb |= (1 << sbb_index)

    def reset_pending_work(self, sbb_index):
        # self.log.info("reset pending work {} for sbb {}".format(self._pending_work_at_sbb, sbb_index))
        self._pending_work_at_sbb &= ~(1 << sbb_index)

    def is_pending_work(self):
        return self._pending_work_at_sbb > 0

    def build_task_list(self):
        # Create a TCP Router socket for comm with other nodes
        # self.router = self.manager.create_socket(socket_type=zmq.ROUTER, name="BM-Router", secure=True)
        self.router = self.manager.create_socket(
            socket_type=zmq.ROUTER,
            name="BM-Router-{}".format(self.verifying_key[-4:]),
            secure=True,
        )
        # self.router.setsockopt(zmq.ROUTER_MANDATORY, 1)  # FOR DEBUG ONLY
        self.router.setsockopt(zmq.IDENTITY, self.verifying_key.encode())
        self.router.bind(port=DELEGATE_ROUTER_PORT, protocol='tcp', ip=self.ip)
        self.tasks.append(self.router.add_handler(self.handle_router_msg))

        # Create ROUTER socket for bidirectional communication with SBBs over IPC
        self.ipc_router = self.manager.create_socket(socket_type=zmq.ROUTER, name="BM-IPC-Router")
        self.ipc_router.setsockopt(zmq.ROUTER_MANDATORY, 1)  # FOR DEBUG ONLY
        self.ipc_router.bind(port=IPC_PORT, protocol='ipc', ip=self.ipc_ip)
        self.tasks.append(self.ipc_router.add_handler(self.handle_ipc_msg, None, True))

        # Create PUB socket to publish new sub_block_contenders to all masters
        # Falcon - is it secure and has a different pub port ??
        #          do we have a corresponding sub at master that handles this properly ?
        self.pub = self.manager.create_socket(
            socket_type=zmq.PUB,
            name="BM-Pub-{}".format(self.verifying_key[-4:]),
            secure=True,
        )
        self.pub.bind(port=DELEGATE_PUB_PORT, protocol='tcp', ip=self.ip)

        self.pub_replacement = self.zmq_ctx.socket(zmq.PUB)

        self.db_state.catchup_mgr = CatchupManager(verifying_key=self.verifying_key,
                                                   signing_key=self.signing_key,
                                                   pub_socket=self.pub,
                                                   router_socket=self.router,
                                                   store_full_blocks=False)

        # Create SUB socket to
        # 1) listen for subblock contenders from other delegates
        # 2) listen for NewBlockNotifications from masternodes
        self.sub = self.manager.create_socket(
            socket_type=zmq.SUB,
            name="BM-Sub-{}".format(self.verifying_key[-4:]),
            secure=True,
        )
        self.sub.setsockopt(zmq.SUBSCRIBE, DEFAULT_FILTER.encode())
        self.sub.setsockopt(zmq.SUBSCRIBE, NEW_BLK_NOTIF_FILTER.encode())

        self.tasks.append(self.sub.add_handler(self.handle_sub_msg))
        self.tasks.append(self._connect_and_process())

    async def _connect_and_process(self):
        # first make sure, we have overlay server ready
        await self._wait_until_ready()

        # Listen to Masternodes over sub and connect router for catchup communication
        for vk in PhoneBook.masternodes:
            self.sub.connect(vk=vk, port=MN_PUB_PORT)
            self.router.connect(vk=vk, port=MN_ROUTER_PORT)

        # now start the catchup
        await self.catchup_db_state()

    async def catchup_db_state(self):
        # do catch up logic here
        await asyncio.sleep(6)  # so pub/sub connections can complete
        assert self.db_state.catchup_mgr, "Expected catchup_mgr initialized at this point"
        self.log.info("Catching up...")

        # Add genesis contracts to state db if needed
        sync_genesis_contracts()

        self.db_state.catchup_mgr.run_catchup()

    def start_sbb_procs(self):
        for i in range(NUM_SB_BUILDERS):
            self.sb_builders[i] = LProcess(target=SubBlockBuilder, name="SBB_Proc-{}".format(i),
                                           kwargs={"ipc_ip": self.ipc_ip, "ipc_port": IPC_PORT,
                                                   "signing_key": self.signing_key, "ip": self.ip,
                                                   "sbb_index": i})

            self.log.info("Starting SBB #{}".format(i))
            self.sb_builders[i].start()

    async def handle_ipc_msg(self, frames):
        self.log.spam("Got msg over ROUTER IPC from a SBB with frames: {}".format(frames))  # TODO delete this
        assert len(frames) == 3, "Expected 3 frames: (id, msg_type, msg_blob). Got {} instead.".format(frames)

        sbb_index = int(frames[0].decode())
        self.log.info('SBBINDEX {}'.format(sbb_index))
        assert sbb_index in self.sb_builders, "Got IPC message with ID {} that is not in sb_builders {}" \
            .format(sbb_index, self.sb_builders)

        mtype_enc = frames[1]
        msg_blob = frames[2]

        msg_type, msg, sender, timestamp, is_verified = Message.unpack_message(mtype_enc, msg_blob)
        if not is_verified:
            self.log.error("Failed to verify the message of type {} from {} at {}. Ignoring it .."
                          .format(msg_type, sender, timestamp))
            return

        if msg_type == MessageType.READY:
            self.log.info('Ready signal received.')
            self.set_sbb_ready()

        elif msg_type == MessageType.PENDING_TRANSACTIONS:
            self.set_pending_work(sbb_index)

        elif msg_type == MessageType.NO_TRANSACTIONS:
            self.reset_pending_work(sbb_index)

        elif msg_type == MessageType.SUBBLOCK_CONTENDER:
            await self._handle_sbc(sbb_index, msg, mtype_enc, msg_blob)

    def handle_sub_msg(self, frames):
        self.log.success('GOT A SUB MESSAGE ON BLOCK MGR')
        self.log.success(frames)

        # Unpack the frames
        msg_filter, msg_type, msg_blob = frames

        msg_type, msg, sender, timestamp, is_verified = Message.unpack_message(msg_type, msg_blob)
        if not is_verified:
            self.log.error("Failed to verify the message of type {} from {} at {}. Ignoring it .."
                          .format(msg_type, sender, timestamp))
            return

        # Process external ready signals
        if msg_type == MessageType.READY:
            # Only allow signals that are sent within 2000 milliseconds to be validated
            if time.time() - timestamp > 2000:
                return

            self._masternodes_ready.add(sender)
            if len(self._masternodes_ready) == PhoneBook.masternode_quorum_min:
                self.send_start_to_sbb()

        # Process block notification messages
        elif msg_type == MessageType.BLOCK_NOTIFICATION:
            self.log.info('Block notification!!')

            self.log.important3("BM got BlockNotification from sender {} with hash {}".format(sender, msg.blockHash.hex()))

            # Process accordingly
            self.handle_block_notification(frames, msg, sender)

    def is_ready_to_start_sub_blocks(self):
        self.start_sub_blocks += 1
        # raghu - wow - who changed this to hard coded 3?
        return self.start_sub_blocks == 3
        
    def send_start_to_sbb(self):
        self.start_sub_blocks += 1
        if self.start_sub_blocks == 3:
            self.send_updated_db_msg()

    def set_catchup_done(self):
        if not self.db_state.is_catchup_done:
            self.db_state.is_catchup_done = True
            self.send_start_to_sbb()

    def set_sbb_ready(self):
        self.sbb_not_ready_count = self.sbb_not_ready_count - 1
        if self.is_sbb_ready():
            self.send_start_to_sbb()
        # log error if count is below 0

    def is_sbb_ready(self):
        return self.sbb_not_ready_count == 0

    def recv_block_data_reply(self, reply):
        # will it block? otherwise, it may not work
        if self.db_state.catchup_mgr.recv_block_data_reply(reply):
            self.set_catchup_done()

    def recv_block_idx_reply(self, sender, reply):
        # will it block? otherwise, it may not work
        if self.db_state.catchup_mgr.recv_block_idx_reply(sender, reply):
            self.set_catchup_done()

    def recv_block_notif(self, block):
        self.db_state.is_catchup_done = False
        # TODO call run_catchup() if catchup_manager is not already catching up
        if self.db_state.catchup_mgr.recv_new_blk_notif(block):
            self.set_catchup_done()

    def handle_router_msg(self, frames):
        sender, msg_type, msg_blob = frames

        msg_type, msg, signer, timestamp, is_verified = Message.unpack_message(msg_type, msg_blob)
        if not is_verified:
            self.log.error("Failed to verify the message of type {} from {} at {}. Ignoring it .."
                          .format(msg_type, signer, timestamp))
            return

        if msg_type == MessageType.BLOCK_INDEX_REPLY:
            self.recv_block_idx_reply(sender, msg)

        elif msg_type == MessageType.BLOCK_DATA:
            self.recv_block_data_reply(msg)

    def _get_new_block_hash(self):
        if not self.db_state.my_sub_blocks.is_quorum():
            return 0
        # first sort the sb result hashes based on sub block index
        sorted_sb_hashes = self.db_state.my_sub_blocks.get_sb_hashes_sorted()

        # append prev block hash

        driver = MetaDataStorage()

        h = hashlib.sha3_256()

        h.update(driver.latest_block_hash)

        for sb_hash in sorted_sb_hashes:
            h.update(sb_hash)

        return h.digest()

    async def _send_sbc(self, mtype_enc: bytes, msg_blob: bytes):
        wait_time = 0
        while not self.is_pending_work() and wait_time < BLOCK_HEART_BEAT_INTERVAL:
            await asyncio.sleep(1)
            wait_time += 1
        # self.log.info("Waited for {} secs. Sending to Masternodes.".format(wait_time))
        # raghu todo - when BM gets a block notification - it should turn off sleep as well as not send in this pub message
        # first sign the message 
        mtype_sgn, msg_sgn = Message.get_message_signed_internal(
                                  signee=self.wallet.verifying_key(),
                                  sign=self.wallet.sign,
                                  msg_type=mtype_enc, msg=msg_blob)
        self.pub.send_msg(filter=DEFAULT_FILTER.encode(),
                          msg_type=mtype_sgn,
                          msg=msg_sgn)

    async def _handle_sbc(self, sbb_index: int, sbc, mtype_enc: bytes, msg_blob: bytes):
        self.log.important("Got SBC with sb-index {} input-hash {}".format(sbc.subBlockIdx, sbc.inputHash.hex()))
        coro = self._send_sbc(mtype_enc, msg_blob)
        fut = asyncio.ensure_future(coro)
        self.db_state.my_sub_blocks.add_sub_block(sbc, fut)

    # TODO make this DRY
    def _send_msg_over_ipc(self, sb_index: int, msg_type, message):
        """
        Convenience method to send a message over IPC router socket to a particular SBB process. Includes a
        frame to identify the type of message
        """
        self.log.spam("Sending msg to sb_index {} with payload {}".format(sb_index, message))

        id_frame = str(sb_index).encode()
        self.ipc_router.send_multipart([id_frame, msg_type, message])

    def _send_fail_block_msg(self, frames):
        for idx in range(NUM_SB_BUILDERS):
            # SIGNAL
            self._send_msg_over_ipc(sb_index=idx, msg_type=frames[1], message=frames[2])

    # make sure block aggregator adds block_num for all notifications?
    def handle_block_notification(self, frames, block, sender: bytes):

        self.log.notice('BM with sender {} being handled'.format(sender))
        self.log.notice("Got block notification for block num {} with hash {}".format(block.blockNum, block.blockHash))

        next_block_num = self.db_state.driver.latest_block_num + 1

        if block.blockNum < next_block_num:
            self.log.info("New block notification with block num {} that is less than or equal to our curr block num {}. "
                          "Ignoring.".format(block.blockNum, self.db_state.driver.latest_block_num))
            return

        if block.blockNum > next_block_num:
            self.log.warning("Current block num {} is behind the block num {} received. Need to run catchup!"
                             .format(self.db_state.driver.latest_block_num, block.blockNum))
            # raghu todo call this below only if it is a new_block_notifi
            self.recv_block_notif(block)     # raghu todo
            return

        is_quorum_met = self.db_state.next_block.add_notification(block, sender, block.blockNum, block.blockHash)

        if is_quorum_met:
            self.log.info("New block quorum met!")
            my_new_block_hash = self._get_new_block_hash()

            self.log.info('New hash {}, recieved hash {}'.format(my_new_block_hash, block.blockHash.hex()))

            if my_new_block_hash == block.blockHash:
                if block.which() == "newBlock":
                    self.db_state.driver.latest_block_num = block.blockNum
                    self.db_state.driver.latest_block_hash = my_new_block_hash

                    # Set the epoch hash if a new epoch has begun
                    if block.blockNum % conf.EPOCH_INTERVAL == 0:
                        self.db_state.driver.latest_epoch_hash = my_new_block_hash

                self.send_updated_db_msg()
                self.nonce_manager.commit_nonces()

                # raghu todo - need to add bgsave for leveldb / redis / ledis if needed here
            else:
                self.log.critical(
                    'BlockNotification hash received is not the same as the one we have!!!\n{}\n{}'.format(
                        my_new_block_hash, block.blockHash))

                # simply forward the block notification. it is input align on sbb
                self._send_fail_block_msg(frames)
                # this can be at sub-block blder level - where it will wait for anothr message only if it is new-block-notif otherwise, it will align input hashes and proceed to make next block
                if block.which() == "newBlock":
                    self.db_state.reset()
                    self.recv_block_notif(block)
                else:
                    self.send_updated_db_msg()

    def send_updated_db_msg(self):
        # first reset my state
        self.db_state.reset()
        self.log.info("Sending MakeNextBlock message to SBBs")

        msg_type, message = Message.get_message_packed(MessageType.MAKE_NEXT_BLOCK)
        for idx in range(NUM_SB_BUILDERS):
            self.ipc_router.send_multipart([str(idx).encode(), msg_type, message])



import election_house

S = Hash()

ELECTION_INTERVAL = datetime.DAYS * 3
VOTING_PERIOD = datetime.DAYS * 1


@construct
def seed(initial_rate=100_000):
    S['rate'] = initial_rate
    reset()


@export
def current_value():
    return S['rate']


@export
def vote(vk, obj):
    assert_vote_is_valid(vk, obj)

    # Check to make sure that there is an election
    if S['in_election']:
        S['votes', vk] = obj

        if now - S['election_start_time'] >= VOTING_PERIOD:
            # Tally votes and set the new value
            result = median(S.all('votes'))
            S['rate'] = result

            reset()
    elif now - S['last_election_end_time'] > ELECTION_INTERVAL:
        # Start the election and set the proper variables
        S['election_start_time'] = now
        S['in_election'] = True
        S['votes', vk] = obj


def assert_vote_is_valid(vk, obj):
    current_rate = S['rate']
    assert type(obj) == int, 'Vote is not an int'
    assert current_rate / 2 <= obj <= current_rate * 2, 'Proposed rate is not within proper boundaries.'

    masternode_policy = election_house.current_value_for_policy(policy='masternodes')

    assert vk in masternode_policy['masternodes'], 'VK is not a masternode!'
    assert S['votes', vk] is None, 'VK already voted!'


def median(vs):
    sorted_votes = sorted(vs)
    index = (len(sorted_votes) - 1) // 2

    if len(sorted_votes) % 2:
        return sorted_votes[index]
    else:
        return (sorted_votes[index] + sorted_votes[index + 1]) / 2


def reset():
    S['last_election_end_time'] = now
    S['in_election'] = False
    S.clear('votes')



from cilantro_ee.messages.block_data.state_update import BlockIndexRequest
from unittest import TestCase


class TestBlockIndexRequest(TestCase):

    def test_serialization(self):
        """
        Tests serialize and from_bytes are inverse operations
        """
        b_hash = 'A' * 64
        b_num = 1260

        sr = BlockIndexRequest.create(block_num=b_num, block_hash=b_hash)
        sr_bin = sr.serialize()

        sr_clone = BlockIndexRequest.from_bytes(sr_bin)

        self.assertEqual(sr, sr_clone)

    def test_serialization_only_block_num(self):
        """
        Tests serialize and from_bytes are inverse operations, with only the block_num field passed in
        """
        b_num = 1260

        sr = BlockIndexRequest.create(block_num=b_num)
        sr_bin = sr.serialize()

        sr_clone = BlockIndexRequest.from_bytes(sr_bin)

        self.assertEqual(sr, sr_clone)
        self.assertTrue(sr_clone.block_hash is None)

    def test_create_0_block(self):
        br = BlockIndexRequest.create(block_num=0)
        self.assertEqual(br.block_num, 0)



from enum import IntEnum, auto, unique

@unique
class MessageBaseType(IntEnum):
    def _generate_next_value_(name, start, count, last_values):
        assert count < 128, "Exceeds the maximum supported message types"
        return 2 * count


class MessageType(MessageBaseType):
    SIGNED_MESSAGE = 0

    MAKE_NEXT_SB = auto()
    COMMIT_CUR_SB = auto()
    DISCORD_AND_ALIGN = auto()

    PENDING_TRANSACTIONS = auto()
    NO_TRANSACTIONS = auto()
    READY = auto()

    SUBBLOCK_CONTENDER = auto()
    BLOCK_INDEX_REQUEST = auto()
    BLOCK_INDEX_REPLY = auto()
    BLOCK_DATA_REQUEST = auto()
    BLOCK_DATA_REPLY = auto()
    BLOCK_DATA = auto()
    BLOCK_NOTIFICATION = auto()
    BURN_INPUT_HASHES = auto()

    TRANSACTION_BATCH = auto()
    # todo - need a better way to handle internal messages inside another message
    TRANSACTION_DATA = auto()
    MERKLE_PROOF = auto()
    TRANSACTION = auto()
    SUBBLOCK = auto()

    LATEST_BLOCK_HEIGHT_REQUEST = auto()
    LATEST_BLOCK_HEIGHT_REPLY = auto()

    LATEST_BLOCK_HASH_REQUEST = auto()
    LATEST_BLOCK_HASH_REPLY = auto()

    BAD_REQUEST = auto()

    IP_FOR_VK_REQUEST = auto()
    IP_FOR_VK_REPLY = auto()



from cilantro_ee.core.messages.notification import *
# import unittest
from unittest import TestCase
import hashlib


class TestBlockNotification(TestCase):

    def test_new_block_notification(self):
        block_hash = 'X3' * 32
        block_num = 32
        first_sb_idx = 4
        block_owners = [ "abc", "def", "pqr"]
        input_hashes = ['AB' * 32, 'BC' * 32, 'C'*64, 'D'*64]

        nbn = BlockNotification.get_new_block_notification(block_num=block_num,
                           block_hash=block_hash, block_owners=block_owners,
                           first_sb_idx=first_sb_idx, input_hashes=input_hashes)

        bn = BlockNotification.unpack_block_notification(nbn)
        self.assertEqual(bn.blockNum, block_num)
        self.assertEqual(len(bn.blockOwners), len(block_owners))
        self.assertEqual(bn.type.which(), "newBlock")
        self.assertNotEqual(bn.type.which(), "emptyBlock")

    def test_empty_block_notification(self):
        prev_hash = 'A' * 64
        block_num = 32
        first_sb_idx = 4
        input_hashes = ['AB' * 32, 'BC' * 32, 'C'*64, 'D'*64]

        if type(prev_hash) == str:
            prev_hash = bytes.fromhex(prev_hash)
        h = hashlib.sha3_256()
        h.update(prev_hash)
        for ih in input_hashes:
            if type(ih) == str:
                ih = bytes.fromhex(ih)
            h.update(ih)
        block_hash = h.digest()

        ebn = BlockNotification.get_empty_block_notification(block_num=block_num,
                           block_hash=block_hash,
                           first_sb_idx=first_sb_idx, input_hashes=input_hashes)

        bn = BlockNotification.unpack_block_notification(ebn)
        self.assertEqual(bn.blockNum, block_num)
        self.assertEqual(bn.blockHash, block_hash)
        self.assertEqual(len(bn.blockOwners), 0)
        self.assertNotEqual(bn.type.which(), "newBlock")
        self.assertEqual(bn.type.which(), "emptyBlock")



import configparser, os
import requests
CIL_CONF_PATH = '/etc/cilantro_ee.conf'

# HOST_IP = None
RESET_DB = False
CONSTITUTION_FILE = None
SSL_ENABLED = None
NONCE_ENABLED = None
STAMPS_ENABLED = False
LOG_LEVEL = None
SEN_LOG_LEVEL = None
SK = None
BOOT_MASTERNODE_IP_LIST = []
BOOT_DELEGATE_IP_LIST = []
BOOTNODES = []
HOST_VK = None
EPOCH_INTERVAL = 1
DEFAULT_DIFFICULTY = 'ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff'
SETUP = False

if not SETUP:
    # Logger is just for debugging
    from cilantro_ee.core.logger.base import get_logger

    log = get_logger("CilantroConf")

    if os.path.exists(CIL_CONF_PATH):
        config = configparser.ConfigParser()
        config.read(CIL_CONF_PATH)
        config = config['DEFAULT']
        CONSTITUTION_FILE = config['constitution_file']
        SK = config['sk']
        RESET_DB = config.getboolean('reset_db')
        SSL_ENABLED = config.getboolean('ssl_enabled')
        NONCE_ENABLED = config.getboolean('nonce_enabled') or False
        STAMPS_ENABLED = config.getboolean('metering')
        LOG_LEVEL = int(config['log_lvl'])
        SEN_LOG_LEVEL = int(config['seneca_log_lvl']) if 'seneca_log_lvl' in config else 0
        BOOT_MASTERNODE_IP_LIST = config['boot_masternode_ips'].split(',')
        BOOT_DELEGATE_IP_LIST = config['boot_delegate_ips'].split(',')
        BOOTNODES = BOOT_MASTERNODE_IP_LIST + BOOT_DELEGATE_IP_LIST
        HOST_IP = requests.get('https://api.ipify.org').text

    SETUP = True



from cilantro_ee.core.logger.base import get_logger
from cilantro_ee.core.utils.worker import Worker

from cilantro_ee.services.storage.state import MetaDataStorage
from cilantro_ee.nodes.catchup import CatchupManager
from cilantro_ee.nodes.masternode.block_contender import BlockContender
from cilantro_ee.services.storage.master import CilantroStorageDriver
from cilantro_ee.constants.zmq_filters import *
from cilantro_ee.constants.ports import MN_ROUTER_PORT, MN_PUB_PORT, DELEGATE_PUB_PORT, SS_PUB_PORT
from cilantro_ee.constants.system_config import *
from cilantro_ee.constants.masternode import *
from cilantro_ee.core.messages.notification import BlockNotification
from cilantro_ee.contracts.sync import sync_genesis_contracts
from cilantro_ee.core.messages.message_type import MessageType
from cilantro_ee.core.messages.message import Message
from cilantro_ee.services.block_fetch import BlockFetcher
import math, asyncio, zmq, time


from cilantro_ee.core.sockets.services import SocketStruct, SubscriptionService


class TransactionBatcherInformer:
    def __init__(self, socket_id):
        pass

    def send_ready(self):
        pass

    def send_burn_input_hashes(self):
        pass


class SubBlockBuilderSubscriber(SubscriptionService):
    pass


class BlockAggregator(Worker):

    def __init__(self, ip, ipc_ip, ipc_port, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.log = get_logger("BlockAggregator[{}]".format(self.verifying_key[:8]))

        assert self.verifying_key in PhoneBook.masternodes, "not a part of VKBook"

        self.ip = ip
        self.ipc_ip = ipc_ip
        self.ipc_port = ipc_port

        self.state = MetaDataStorage()

        self.curr_block = BlockContender()

        self.pub, self.sub, self.router, self.ipc_router = None, None, None, None  # Set in build_task_list

        self.catchup_manager = None  # This gets set at the end of build_task_list once sockets are created
        self.timeout_fut = None

        self._is_catchup_done = False

        self.min_quorum = PhoneBook.delegate_quorum_min
        self.max_quorum = PhoneBook.delegate_quorum_max
        self.cur_quorum = 0

        self.my_mn_idx = PhoneBook.masternodes.index(self.verifying_key)
        self.my_sb_idx = self.my_mn_idx % NUM_SB_BUILDERS

        self.curr_block_hash = self.state.get_latest_block_hash()

        self.driver = CilantroStorageDriver(key=self.signing_key)

        last_block = self.driver.get_last_n(1, CilantroStorageDriver.INDEX)[0]
        latest_hash = last_block.get('blockHash')
        latest_num = last_block.get('blockNum')

        self.block_fetcher = BlockFetcher(wallet=self.wallet,
                                          ctx=self.zmq_ctx,
                                          blocks=self.driver,
                                          state=self.state)

        self.state.latest_block_num = latest_num
        self.state.latest_block_hash = latest_hash

        self.run()

    def run(self):
        self.log.info("Block Aggregator starting...")
        self.build_task_list()

        self.loop.run_until_complete(asyncio.gather(*self.tasks))

    def build_task_list(self):
        # Create ROUTER socket for communication with batcher over IPC
        # DEALER SOCKET...
        self.ipc_router = self.manager.create_socket(socket_type=zmq.ROUTER, name="BA-IPC-Router")
        self.ipc_router.bind(port=self.ipc_port, protocol='ipc', ip=self.ipc_ip)

        # SubscriptionService
        self.sub = self.manager.create_socket(
            socket_type=zmq.SUB,
            name="BA-Sub",
            secure=True,
        )
        self.sub.setsockopt(zmq.SUBSCRIBE, BLOCK_IDX_REQ_FILTER.encode())
        self.sub.setsockopt(zmq.SUBSCRIBE, DEFAULT_FILTER.encode())
        self.sub.setsockopt(zmq.SUBSCRIBE, NEW_BLK_NOTIF_FILTER.encode())

        # Regular Publish socket
        self.pub = self.manager.create_socket(
            socket_type=zmq.PUB,
            name="BA-Pub",
            secure=True,
        )
        self.pub.bind(ip=self.ip, port=MN_PUB_PORT)

        ### AsyncServer?
        self.router = self.manager.create_socket(
            socket_type=zmq.ROUTER,
            name="BA-Router",
            secure=True,
        )
        self.router.setsockopt(zmq.IDENTITY, self.verifying_key.encode())
        self.router.bind(ip=self.ip, port=MN_ROUTER_PORT)

        self.tasks.append(self.sub.add_handler(self.handle_sub_msg))
        self.tasks.append(self.router.add_handler(self.handle_router_msg))

        self.catchup_manager = CatchupManager(verifying_key=self.verifying_key, signing_key=self.signing_key, pub_socket=self.pub,
                                              router_socket=self.router, store_full_blocks=True)

        self.tasks.append(self._connect_and_process())

    async def _connect_and_process(self):
        # first make sure, we have overlay server ready
        await self._wait_until_ready()

        self.log.info('connecting to masters: {}'.format(PhoneBook.masternodes))
        self.log.info('connecting to delegates: {}'.format(PhoneBook.delegates))

        # Listen to masters for _new block notifs and state update requests from masters/delegates
        for vk in PhoneBook.masternodes:
            if vk != self.verifying_key:
                self.sub.connect(vk=vk, port=MN_PUB_PORT)
                # self.router.connect(vk=vk, port=MN_ROUTER_PORT)  # we don't want 2 simultaneous look ups @ overlay server

        # Listen to delegates for sub block contenders and state update requests
        for vk in PhoneBook.delegates:
            self.sub.connect(vk=vk, port=DELEGATE_PUB_PORT)
            # I dont think we to connect to delegates to router here as delegates are already connecting
            # in BlockManager --davis
            # self.router.connect(vk=vk, port=DELEGATE_ROUTER_PORT)

        for vk in PhoneBook.schedulers + PhoneBook.notifiers:
            self.sub.connect(vk=vk, port=SS_PUB_PORT)

        # Listen to masters for new block notifs and state update requests from masters/delegates
        for vk in PhoneBook.masternodes:
            if vk != self.verifying_key:
                self.router.connect(vk=vk, port=MN_ROUTER_PORT)

        # we just connected to other nodes, let's chill a bit to give time for those connections form !!!

        # Do a dealer / router socket pair here instead.
        self.log.info("Sleeping before triggering catchup...")
        await asyncio.sleep(8)

        num_delegates_joined = self.manager.get_and_reset_num_delegates_joined()
        # assert num_delegates_joined >= self.min_quorum, "Don't have minimum quorum"
        if num_delegates_joined >= self.max_quorum:
            self.cur_quorum = self.max_quorum
        else:
            cq = math.ceil(9 * num_delegates_joined / 10)
            self.cur_quorum = max(cq, self.min_quorum)

        # now start the catchup
        # sync_genesis_contracts()
        # await self.block_fetcher.sync()

        await self._trigger_catchup()

    async def _trigger_catchup(self):
        self.log.info("Triggering catchup")
        # Add genesis contracts to state db if needed
        sync_genesis_contracts()

        self.catchup_manager.run_catchup()

### SUB MESSAGE LOOP SHOULD BE ASYNC
    def handle_sub_msg(self, frames):
        msg_filter, msg_type, msg_blob = frames
        self.log.success(len(frames))

        msg_type, msg, sender, timestamp, is_verified = Message.unpack_message(msg_type, msg_blob)
        if not is_verified:
            self.log.error("Failed to verify the message of type {} from {} at {}. Ignoring it .."
                          .format(msg_type, sender, timestamp))
            return

        # Move this socket to where the communication is happening (IE: the blockserver?)
        if msg_type == MessageType.BLOCK_INDEX_REQUEST:
            self.catchup_manager.recv_block_idx_req(msg)

        elif msg_type == MessageType.SUBBLOCK_CONTENDER:
            if not self.catchup_manager.is_catchup_done():
                self.log.info("Got SBC, but i'm still catching up. Ignoring: <{}>".format(msg))
            else:
                # todo needs to handle this ??
                # signature = subblock_capnp.MerkleProof.from_bytes_packed(msg.signature)
                self.recv_sub_block_contender(sender, msg)

        # Process block notification messages
        elif msg_type == MessageType.BLOCK_NOTIFICATION:
            self.log.important3(
                "BlockAGG got BlockNotification from sender {} with hash {}"
                .format(sender, msg.blockHash))
            # Process accordingly
            self.recv_new_block_notif(sender, msg)

    def _set_catchup_done(self):
        if not self._is_catchup_done:
            self._is_catchup_done = True
            self.curr_block_hash = self.state.get_latest_block_hash()
            self.curr_block.reset()

            mtype, msg = Message.get_message_packed(MessageType.READY)
            self.ipc_router.send_multipart([b'0', mtype, msg])

            time.sleep(3)

            # Construct a cryptographically signed message of the current time such that the receiver can verify it
            mtype, msg = Message.get_signed_message_packed(
                                    wallet=self.wallet,
                                    msg_type=MessageType.READY)

### Send signed READY signal on pub
            self.log.success('READY SIGNAL SENT TO SUBS')
            self.pub.send_msg(msg=msg, msg_type=mtype,
                              filter=DEFAULT_FILTER.encode())

    # Can we put this exclusively in Catchup manager?
    def handle_router_msg(self, frames):
        self.log.info('got message on block agg with frames {}'.format(frames))

        sender, msg_type, msg_blob = frames

        # need to distinguish between sender (remote socket identity vs. message signer which is vk)
        msg_type, msg, signee, timestamp, is_verified = Message.unpack_message(msg_type, msg_blob)
        if not is_verified:
            self.log.error("Failed to verify the message of type {} from {} at {}. Ignoring it .."
                          .format(msg_type, sender, timestamp))
            return

        # When does this happen?
        if msg_type == MessageType.BLOCK_INDEX_REPLY:
            if self.catchup_manager.recv_block_idx_reply(sender, msg):
                self._set_catchup_done()

        elif msg_type == MessageType.BLOCK_DATA_REQUEST:
            self.catchup_manager.recv_block_data_req(sender, msg)

        elif msg_type == MessageType.BLOCK_DATA:
            if self.catchup_manager.recv_block_data_reply(msg):
                self._set_catchup_done()


    # Most likely testable
    def recv_sub_block_contender(self, sender_vk: str, sbc):
        self.log.debugv("Received a sbc from sender {} with result hash {} and input hash {}"
                        .format(sender_vk, sbc.resultHash, sbc.inputHash))

        added_first_sbc = self.curr_block.add_sbc(sender_vk, sbc)

        if added_first_sbc:
            self.log.debug("First SBC receiver for prev block hash {}! Scheduling timeout".format(self.curr_block_hash))
            self.timeout_fut = asyncio.ensure_future(self.schedule_block_timeout())

        if self.curr_block.is_consensus_reached() or \
           self.curr_block.get_current_quorum_reached() >= self.cur_quorum:
            self.log.spam("currnt quorum {} actual quorum {} max quorum {}".
                          format(self.cur_quorum, self.curr_block.get_current_quorum_reached(), self.max_quorum))
            self.log.success("Consensus reached for prev hash {} (is_empty={})"
                             .format(self.curr_block_hash, self.curr_block.is_empty()))
            self.store_full_block()
            num_delegates_joined = self.manager.get_and_reset_num_delegates_joined()
            self.cur_quorum = min(self.cur_quorum + num_delegates_joined, self.max_quorum)
            return

        if not self.curr_block.is_consensus_possible():
            self.log.critical("Consensus not possible for prev block hash {}! Sending failed block notif".format(self.curr_block_hash))
            self.send_fail_block_notif()
            self._reset_curr_block()
        else:
            self.log.debugv("Consensus not reached yet.")

    # Very testable
    def store_full_block(self):
        sb_data = self.curr_block.get_sb_data()

        self.log.info(sb_data)

        if self.curr_block.is_empty():
            self.log.info("Got consensus on empty block with prev hash {}! Sending skip block notification".format(self.curr_block_hash))
            time.sleep(10)
            self.send_skip_block_notif(sb_data)

        else:
            # TODO wrap storage in try/catch. Add logic for storage failure
            self.log.info("Storing a block: {}".format(self.curr_block_hash))

            #try:
            block_data = self.driver.store_block(sb_data)
            self.log.debug(block_data)

            assert block_data['prevBlockHash'] == self.curr_block_hash, \
                "Current block hash {} does not match StorageDriver previous block hash {}"\
                .format(self.curr_block_hash, block_data['prevBlockHash'])

            self.curr_block_hash = block_data['blockHash']
            self.state.update_with_block(block_data)

            self.log.success2("STORED BLOCK WITH HASH {}".format(block_data['blockHash']))

            self.send_new_block_notif(block_data)

            #except Exception as e:
            #    self.log.error(str(e))

        # TODO
        # @tejas yo why does this assertion not pass? The storage blocks is NOT updating its block hash after storing!
        # assert StorageDriver.get_latest_block_hash() == self.state.get_latest_block_hash(), \
        #     "StorageDriver latest block hash {} does not match StateDriver latest hash {}" \
        #         .format(StorageDriver.get_latest_block_hash(), self.state.get_latest_block_hash())

        self._reset_curr_block()


    def send_block_notif(self, msg_type, **kwargs):
        mn_idx = kwargs.get('firstSbIdx') + self.my_sb_idx
        if mn_idx == self.my_mn_idx:
            input_hashes = kwargs.get('inputHashes')
            my_input_hashes = input_hashes[self.my_sb_idx]
            self.log.info("Need to burn input bags with hash(es) {}".format(my_input_hashes))
            bih_mtype, bih = Message.get_message_packed(MessageType.BURN_INPUT_HASHES, inputHashes=my_input_hashes)
            self.ipc_router.send_multipart([b'0', bih_mtype, bih])

        mtype, bn = Message.get_signed_message_packed(
                             wallet=self.wallet,
                             msg_type=msg_type, **kwargs)

        # clean up filters for different block notifications - unify it under BLK_NOTIF_FILTER ?
        self.pub.send_msg(filter=DEFAULT_FILTER.encode(), msg_type=mtype, msg=bn)


    def send_new_block_notif(self, block_data):
        # sleep a bit so slower nodes don't have to constantly use catchup mgr 
        time.sleep(0.1)

        # SEND NEW BLOCK NOTIFICATION on pub
        self.send_block_notif(MessageType.BLOCK_NOTIFICATION,
                              blockNum=block_data['blockNum'],
                              blockHash=block_data['blockHash'],
                              blockOwners=block_data['blockOwners'],
                              firstSbIdx=block_data['subBlocks'][0].subBlockIdx,
                              inputHashes=[[sb.inputHash] for sb in block_data['subBlocks']],
                              newBlock=None)
        self.log.info('Published new block notif with hash "{}" and block num {}'
                      .format(block_data['blockHash'], block_data['blockNum']))

    def send_skip_block_notif(self, sub_blocks):
        # assert that sub_blocks are sorted by subBlockIdx
        last_hash = self.state.latest_block_hash
        block_num = self.state.latest_block_num + 1
        input_hashes = [[sb.inputHash] for sb in sub_blocks]
        block_hash = BlockNotification.get_block_hash(last_hash, input_hashes)

        self.send_block_notif(MessageType.BLOCK_NOTIFICATION,
                              blockNum=block_num,
                              blockHash=block_hash,
                              blockOwners=[],
                              firstSbIdx=sub_blocks[0].subBlockIdx,
                              inputHashes=input_hashes,
                              emptyBlock=None)

        self.log.debugv("Published empty block notification for hash {} num {}"
                        .format(block_hash, block_num))

    def send_fail_block_notif(self):
        last_hash = self.state.latest_block_hash
        block_num = self.state.latest_block_num + 1
        input_hashes = self.curr_block.get_input_hashes_sorted()
        block_hash = BlockNotification.get_block_hash(last_hash, input_hashes)
        first_sb_idx = self.curr_block.get_first_sb_idx_sorted()

        self.send_block_notif(MessageType.BLOCK_NOTIFICATION,
                              blockNum=block_num,
                              blockHash=block_hash,
                              blockOwners=[],
                              firstSbIdx=first_sb_idx,
                              inputHashes=input_hashes,
                              failedBlock=None)

        self.log.debugv("Published failed block notification for hash {} num {}"
                        .format(block_hash, block_num))

    # Can we put this on CatchupManager / BlockServer instead?
    def recv_new_block_notif(self, sender_vk: str, notif):
        self.log.debugv("MN got new block notification: {}".format(notif))

        blocknum = notif.blockNum

        if (blocknum > self.state.latest_block_num + 1) and \
           (notif.type.which() == "newBlock"):
            self.log.info("Block num {} on NBC does not match our block num {}! Triggering catchup".format(notif.block_num, self.state.latest_block_num))

            # Call intermediate sync on Block Fetcher
            self.catchup_manager.recv_new_blk_notif(notif)
        else:
            self.log.debugv("Block num on NBC is LTE that ours. Ignoring")

    def recv_fail_block_notif(self, sender_vk: str, notif):
        self.log.debugv("MN got fail block notification: {}".format(notif))
        # TODO implement

    # Probably testable
    async def schedule_block_timeout(self):
        try:
            elapsed = 0

            while elapsed < BLOCK_PRODUCTION_TIMEOUT:
                await asyncio.sleep(BLOCK_TIMEOUT_POLL)
                elapsed += BLOCK_TIMEOUT_POLL

            self.log.critical("Block timeout of {}s reached for block hash {}!"
                              .format(BLOCK_PRODUCTION_TIMEOUT, self.curr_block_hash))
            new_quorum = self.curr_block.get_current_quorum_reached()
            if new_quorum >= self.min_quorum and new_quorum >= (9 * self.cur_quorum // 10):
                self.log.warning("Reducing consensus quorum from {} to {}".
                          format(self.curr_block.get_current_quorum_reached(), new_quorum))
                self.store_full_block()
                self.cur_quorum = new_quorum
            else:
                self.log.debugv("sending fail block notif")
                self.send_fail_block_notif()
            self.curr_block.reset()
            num_delegates_joined = self.manager.get_and_reset_num_delegates_joined()
            self.cur_quorum = min(self.cur_quorum + num_delegates_joined, self.max_quorum)
        except asyncio.CancelledError:
            pass

    def _reset_curr_block(self):
        self.curr_block.reset()
        self.log.debugv("Canceling block timeout")
        if self.timeout_fut and not self.timeout_fut.done():
            self.timeout_fut.cancel()





"""
BlockMetaData is just a subclass of BlockData, with no additional fields, so tests should be covered by test_block_data

-- davis
"""

# from unittest import TestCase
# from unittest import mock
#
# from cilantro_ee.core.crypto import wallet
# from cilantro_ee.messages.block_data.block_data import BlockData
# from cilantro_ee.messages.consensus.merkle_signature import MerkleSignature
#
# from cilantro_ee.constants.testnet import TESTNET_MASTERNODES
# TEST_SK, TEST_VK = TESTNET_MASTERNODES[0]['sk'], TESTNET_MASTERNODES[0]['vk']
#
#
# class TestBlockMetaData(TestCase):
#
#     @mock.patch("cilantro_ee.messages.block_data.block_metadata.NUM_SB_PER_BLOCK", 2)
#     def test_create(self):
#         from cilantro_ee.messages.block_data.block_metadata import BlockMetaData  # avoid cyclic imports
#
#         prev_b_hash = 'A' * 64
#         input_hashes = ['B' * 64, 'C' * 64]
#         roots = ['D' * 64, 'E' * 64]
#         timestamp = 1337
#         block_num = 12
#         b_hash = BlockData.compute_block_hash(sbc_roots=roots, prev_block_hash=prev_b_hash)
#
#         block_owners = [TEST_VK]
#
#         block_meta = BlockMetaData.create(block_hash=b_hash, merkle_roots=roots, input_hashes=input_hashes,
#                                           prev_block_hash=prev_b_hash, block_owners=block_owners, timestamp=timestamp,
#                                           block_num=block_num)
#
#         self.assertEqual(block_meta.prev_block_hash, prev_b_hash)
#         self.assertEqual(block_meta.input_hashes, input_hashes)
#         self.assertEqual(block_meta.merkle_roots, roots)
#         self.assertEqual(block_meta.timestamp,timestamp)
#         self.assertEqual(block_meta.block_hash, b_hash)
#         self.assertEqual(block_meta.block_num, block_num)
#         self.assertEqual(block_meta.block_owners, block_owners)
#
#     @mock.patch("cilantro_ee.messages.block_data.block_metadata.NUM_SB_PER_BLOCK", 2)
#     def test_clone(self):
#         from cilantro_ee.messages.block_data.block_metadata import BlockMetaData  # avoid cyclic imports
#
#         prev_b_hash = 'A' * 64
#         input_hashes = ['B' * 64, 'C' * 64]
#         roots = ['D' * 64, 'E' * 64]
#         timestamp = 1337
#         block_num = 12
#         b_hash = BlockData.compute_block_hash(sbc_roots=roots, prev_block_hash=prev_b_hash)
#
#         block_owners = [TEST_VK]
#
#         block_meta = BlockMetaData.create(block_hash=b_hash, merkle_roots=roots, input_hashes=input_hashes,
#                                           prev_block_hash=prev_b_hash, block_owners=block_owners, timestamp=timestamp,
#                                           block_num=block_num)
#         clone = BlockMetaData.from_bytes(block_meta.serialize())
#
#         self.assertEqual(block_meta, clone)
#
#     @mock.patch("cilantro_ee.messages.block_data.block_metadata.NUM_SB_PER_BLOCK", 2)
#     def test_from_block_data(self):
#         # TODO implement
#         pass





import cilantro_ee
from cilantro_ee.core.crypto import wallet
from pymongo import MongoClient, DESCENDING
from configparser import ConfigParser
from cilantro_ee.core.logger.base import get_logger
from bson.objectid import ObjectId
from collections import defaultdict
from typing import List
from cilantro_ee.constants.system_config import *
from cilantro_ee.services.storage.vkbook import PhoneBook
from cilantro_ee.constants import system_config

import hashlib

REPLICATION = 3             # TODO hard coded for now needs to change
GENESIS_HASH = '0' * 64
OID = '5bef52cca4259d4ca5607661'


class StorageSet:
    def __init__(self, user, password, port, database, collection_name):
        self.uri = 'mongodb://{}:{}@localhost:{}/{}?authSource=admin&maxPoolSize=1'.format(
            user,
            password,
            port,
            database
        )

        self.client = MongoClient(self.uri)
        self.db = self.client.get_database()
        self.collection = self.db[collection_name]

    def flush(self):
        self.client.drop_database(self.db)


class MasterStorage:
    BLOCK = 0
    INDEX = 1
    TX = 2

    def __init__(self, config_path=cilantro_ee.__path__[0]):
        # Setup configuration file to read constants
        self.config_path = config_path

        self.config = ConfigParser()
        self.config.read(self.config_path + '/services/storage/mn_db_conf.ini')

        user = self.config.get('MN_DB', 'username')
        password = self.config.get('MN_DB', 'password')
        port = self.config.get('MN_DB', 'port')

        block_database = self.config.get('MN_DB', 'mn_blk_database')
        index_database = self.config.get('MN_DB', 'mn_index_database')
        tx_database = self.config.get('MN_DB', 'mn_tx_database')

        self.blocks = StorageSet(user, password, port, block_database, 'blocks')
        self.indexes = StorageSet(user, password, port, index_database, 'index')
        self.txs = StorageSet(user, password, port, tx_database, 'tx')

        if self.get_block(0) is None:
            self.put({
                'blockNum': 0,
                'blockHash': '0' * 64,
                'blockOwners': ['0' * 64]
            }, MasterStorage.BLOCK)

            self.put({
                'blockNum': 0,
                'blockHash': '0' * 64,
                'blockOwners': ['0' * 64]
            }, MasterStorage.INDEX)

    def q(self, v):
        if isinstance(v, int):
            return {'blockNum': v}
        return {'blockHash': v}

    def get_block(self, v=None):
        if v is None:
            return None

        q = self.q(v)
        block = self.blocks.collection.find_one(q)

        if block is not None:
            block.pop('_id')

        return block

    def put(self, data, collection=BLOCK):
        if collection == MasterStorage.BLOCK:
            _id = self.blocks.collection.insert_one(data)
        elif collection == MasterStorage.INDEX:
            _id = self.indexes.collection.insert_one(data)
        elif collection == MasterStorage.TX:
            _id = self.txs.collection.insert_one(data)
        else:
            return False

        return _id is not None

    def get_last_n(self, n, collection=INDEX):
        if collection == MasterStorage.BLOCK:
            c = self.blocks
        elif collection == MasterStorage.INDEX:
            c = self.indexes
        else:
            return None

        block_query = c.collection.find({}, {'_id': False}).sort(
            'blockNum', DESCENDING
        ).limit(n)

        blocks = [block for block in block_query]

        if len(blocks) > 1:
            first_block_num = blocks[0].get('blockNum')
            last_block_num = blocks[-1].get('blockNum')

            assert first_block_num > last_block_num, "Blocks are not descending."

        return blocks

    def get_owners(self, v):
        q = self.q(v)
        index = self.indexes.collection.find_one(q)

        if index is None:
            return index

        owners = index.get('blockOwners')

        return owners

    def get_index(self, v):
        q = self.q(v)
        block = self.indexes.collection.find_one(q)

        if block is not None:
            block.pop('_id')

        return block

    def get_tx(self, h):
        tx = self.txs.collection.find_one({'tx_hash': h})

        if tx is not None:
            tx.pop('_id')

        return tx

#    def put_tx_map(self, block: BlockData):
#        m = block.get_tx_hash_to_merkle_leaf()
#        blk_id = block.block_num
#
#        for entry in m:
#            entry['block'] = blk_id
#            self.txs.collection.insert_one(entry)

    def drop_collections(self):
        self.blocks.flush()
        self.indexes.flush()


class DistributedMasterStorage(MasterStorage):
    def __init__(self, key, distribute_writes=False, config_path=cilantro_ee.__path__[0], vkbook=system_config.PhoneBook):
        super().__init__(config_path=config_path)

        self.distribute_writes = distribute_writes
        self.vkbook = vkbook
        self.sk = key
        self.vk = wallet.get_vk(self.sk)

        self.test_hook = self.config.get('MN_DB', 'test_hook')
        self.mn_id = int(self.config.get('MN_DB', 'mn_id'))
        self.rep_factor = int(self.config.get('MN_DB', 'replication'))
        self.active_masters = int(self.config.get('MN_DB', 'total_mn'))
        self.quorum_needed = int(self.config.get('MN_DB', 'quorum'))

    def get_master_set(self):
        if self.test_hook is True:
            return self.active_masters
        else:
            self.active_masters = len(self.vkbook.masternodes)
            return self.active_masters

    def set_mn_id(self, vk):
        if self.test_hook is True:
            return self.mn_id

        # this should be rewritten to just pull from Phonebook because it's dynamic now

        for i in range(self.get_master_set()):
            if self.vkbook.masternodes[i] == vk:
                self.mn_id = i
                return True
            else:
                self.mn_id = -1
                return False

    def rep_pool_sz(self):
        if self.active_masters < self.rep_factor:
            return -1

        self.active_masters = self.get_master_set()
        pool_sz = round(self.active_masters / self.rep_factor)
        return pool_sz

    def build_wr_list(self, curr_node_idx=0, jump_idx=1):
        # Use slices to make this a one liner
        tot_mn = len(self.vkbook.masternodes)
        mn_list = []

        # if quorum req not met jump_idx is 0 wr on all active nodes
        if jump_idx == 0:
            return self.vkbook.masternodes

        while curr_node_idx < tot_mn:
            mn_list.append(self.vkbook.masternodes[curr_node_idx])
            curr_node_idx += jump_idx

        return mn_list

    @staticmethod
    def index_from_block(b, nodes=[]):
        assert len(nodes) > 0, 'Must have at least one block owner!'

        index = {'blockNum': b.get('blockNum'),
                 'blockHash': b.get('blockHash'),
                 'blockOwners': nodes}

        assert index['blockHash'] is not None and index['blockNum'] is not None, 'Block hash and number' \
                                                                                 'must be provided!'

        return index

    def evaluate_wr(self, entry=None, node_id=None):
        """
        Function is used to check if currently node is suppose to write given entry

        :param entry: given block input to be stored
        :param node_id: master id None is default current master, if specified is for catch up case
        :return:
        """

        if entry is None:
            return False

        pool_sz = self.rep_pool_sz()
        mn_idx = self.mn_id % pool_sz
        writers = entry.get('blockNum') % pool_sz

        # TODO
        # need gov here to check if given node is voted out

        if node_id is not None:
            mn_idx = node_id % pool_sz  # overwriting mn_idx
            if mn_idx == writers:
                return True
            else:
                return False

        # always write if active master bellow threshold

        if self.active_masters < self.quorum_needed:
            self.put(entry, self.BLOCK)
            mn_list = self.build_wr_list(curr_node_idx=self.mn_id, jump_idx=0)
            index = self.index_from_block(entry, nodes=mn_list)
            return self.put(index, self.INDEX)

        if mn_idx == writers:
            self.put(entry, self.BLOCK)

        # build list of mn_sign of master nodes updating index db
        mn_list = self.build_wr_list(curr_node_idx=writers, jump_idx=pool_sz)
        assert len(mn_list) > 0, "block owner list cannot be empty - dumping list -> {}".format(mn_list)

        # create index records and update entry
        index = self.index_from_block(entry, nodes=mn_list)
        return self.put(index, self.INDEX)

    def update_index(self, block, nodes):
        index = self.index_from_block(block, nodes=nodes)
        return self.put(index, MasterStorage.INDEX)

    def put(self, data, collection=MasterStorage.BLOCK):
        super().put(data=data, collection=collection)


class CilantroStorageDriver(DistributedMasterStorage):
    def __init__(self, key, distribute_writes=False, config_path=cilantro_ee.__path__[0], vkbook=system_config.PhoneBook):
        self.state_id = ObjectId(OID)
        self.log = get_logger("StorageDriver")

        self.block_index_delta = defaultdict(dict)
        self.send_req_blk_num = 0

        super().__init__(key, distribute_writes=distribute_writes, config_path=config_path, vkbook=vkbook)

    def store_block(self, sub_blocks):
        last_block = self.get_last_n(1, self.INDEX)[0]

        last_hash = last_block.get('blockHash')
        current_block_num = last_block.get('blockNum') + 1

        hashes = [subblock.merkleRoot for subblock in sub_blocks]

        for s in sub_blocks:
            self.log.success(type(s))

        if type(last_hash) == str:
            last_hash = bytes.fromhex(last_hash)

        h = hashlib.sha3_256()
        h.update(last_hash)

        for _hash in hashes:
            h.update(_hash)

        block_hash = h.digest()

        block_dict = {
            'blockHash': block_hash,
            'blockNum': current_block_num,
            'blockOwners': [m for m in PhoneBook.masternodes],
            'prevBlockHash': last_hash,
            'subBlocks': [s for s in sub_blocks]
        }

        # Serialize the sub block for mongo
        block_dict['subBlocks'] = [s.to_bytes_packed() for s in block_dict['subBlocks']]

        #if not self.distribute_writes:
        #    block_data = BlockData.create(block_hash, last_hash, PhoneBook.masternodes, current_block_num, sub_blocks)

        successful_storage = self.evaluate_wr(entry=block_dict)

        assert successful_storage is None or successful_storage is True, 'Write failure.'
        block_dict['subBlocks'] = [s for s in sub_blocks]

        return block_dict

    def get_transactions(self, tx_hash):
        txs = self.get_tx(tx_hash)

        if txs is None:
            return None

        block_num = txs.get('block')
        leaf = txs.get('tx_leaf')

        block = self.get_block(block_num)
        sub_blocks = block.get('subBlocks')

        for i in range(0, system_config.NUM_SB_PER_BLOCK):
            leaves = sub_blocks[i].get('merkleLeaves')

            try:
                tx_idx = leaves.index(leaf)
            except ValueError:
                tx_idx = -1

            if tx_idx >= 0:
                tx_dump = sub_blocks[i].get('transactions')
                return tx_dump[tx_idx]

        return None



from cilantro_ee.services.storage.vkbook import PhoneBook, VKBook
from cilantro_ee.services.storage.state import MetaDataStorage
from cilantro_ee.core.logger.base import get_logger
from cilantro_ee.core.containers.merkle_tree import MerkleTree
from cilantro_ee.constants.system_config import *
from cilantro_ee.utils.hasher import Hasher

from cilantro_ee.core.messages.message_type import MessageType
from cilantro_ee.core.messages.message import Message
from cilantro_ee.core.crypto.wallet import _verify

from collections import defaultdict
from typing import List
import time
import hashlib


class SBCSenderSignerMismatchError(Exception):
    pass

class SBCIndexMismatchError(Exception):
    pass


class SBCInvalidSignatureError(Exception):
    pass


class SBCTransactionNotInContenderError(Exception):
    pass


class SBCBlockHashMismatchError(Exception):
    pass


class SBCMerkleLeafVerificationError(Exception):
    pass


class SBCIndexGreaterThanPossibleError(Exception):
    pass

class SubBlockGroup:
    def __init__(self, sb_idx: int, curr_block_hash: str, contacts: VKBook=PhoneBook):
        self.sb_idx, self.curr_block_hash = sb_idx, curr_block_hash
        self.log = get_logger("SBGroup[{}]".format(self.sb_idx))

        self.rh = defaultdict(set)  # mapping of result_hash: set of SubBlockContenders

        self.transactions = {}  # tx_hash: TransactionData
        self.sender_to_sbc = {}  # map of sender_vk: SubBlockContender

        self.contacts = contacts

        self.min_quorum = self.contacts.delegate_quorum_min
        self.max_quorum = self.contacts.delegate_quorum_max

        self.best_rh = None  # The result hash with the most votes so far

    def is_consensus_possible(self) -> bool:
        num_votes = 0
        for rh in self.rh:
            votes_for_rh = len(self.rh[rh])
            if votes_for_rh >= self.max_quorum:  # Consensus reached on this sub block
                return True
            num_votes += votes_for_rh

        remaining_votes = len(self.contacts.delegates) - num_votes
        leading_rh = len(self.rh[self.best_rh])

        if leading_rh + remaining_votes < self.max_quorum:
            self.log.fatal("Consensus impossible for SB index {}!\n"
                           "consensus: {},\n"
                           "result hashes: {}".format(self.sb_idx, leading_rh, self.rh))
            return False

        return True

    def get_sb(self):
        #assert self.is_consensus_reached(), "Consensus must be reached to get a SubBlock!"

        # Paranoid dev checks
        # TODO make sure all merkle leaves are the same, and all result hashes are the same for self.rh[merkle_root],
        # all sb_idx matches ours, all input hashes are the same

        merkle_root = self.best_rh
        contenders = self.rh[merkle_root]

        sigs = [c.signature for c in contenders]

        # Get a contender from the set. This presumes that the contenders have identical data, which they should.
        # contenders should be grouped based on prevBlockHash and resultHash equality - then they will have identical data - raghu
        contender = contenders.pop()
        contenders.add(contender)

        leaves = contender.merkleLeaves
        input_hash = contender.inputHash

        txs = self._get_ordered_transactions()
        transactions = [Message.unpack_message_internal(MessageType.TRANSACTION_DATA, tx) for tx in txs]

        # looks like sbc has txns packed while sb will have them as unpacked. Need to eliminate these inconsistencies for better performance - raghu
        _, sb = Message.get_message(
                   MessageType.SUBBLOCK, merkleRoot=merkle_root,
                   signatures=sigs, merkleLeaves=[leaf for leaf in leaves],
                   subBlockIdx=self.sb_idx, inputHash=input_hash,
                   transactions=transactions)
   
        return sb

    def is_consensus_reached(self) -> bool:
        cons_reached = len(self.rh[self.best_rh]) >= self.max_quorum

        return cons_reached

    def get_current_quorum_reached(self) -> int:
        # If the best result is still less than the minimum required quorum, return zero
        if len(self.rh[self.best_rh]) < self.min_quorum:
            return 0

        # If the best result is more than the max quorum, return the max quorum
        if len(self.rh[self.best_rh]) >= self.max_quorum:
            return self.max_quorum

        # Otherwise, calculate the total number of votes recieved at this point in time
        num_votes = 0

        for rh in self.rh:
            votes_for_rh = len(self.rh[rh])
            num_votes += votes_for_rh

        # Get the number of votes of the most popular result
        leading_rh = len(self.rh[self.best_rh])

        # The quorum is reduced if the most popular result is 90% of the current votes
        is_reduced_quorum = leading_rh >= (9 * num_votes // 10)

        # If this is the case, return the number for the most popular votes
        return leading_rh if is_reduced_quorum else 0

    def get_input_hashes(self) -> list:
        s = set()

        for sbc in self.sender_to_sbc.values():
            s.add(sbc.inputHash)

        return list(s)

    def is_empty(self):
        return len(self._get_merkle_leaves()) == 0

    def _get_merkle_leaves(self) -> list:
        if self.best_rh is None:
            return []

        # All merkle leaves should be the same, so just chose any contender from the set
        return next(iter(self.rh[self.best_rh])).merkleLeaves

    def _get_ordered_transactions(self):
        #assert self.is_consensus_reached(), "Must be in consensus to get ordered transactions"  # TODO remove

        # ... Doesn't this return tx's for ALL SBC? WTF IS GOING ON HERE....
        # return [self.transactions[tx_hash] for tx_hash in self._get_merkle_leaves()]
        return next(iter(self.rh[self.best_rh])).transactions

    def add_sbc(self, sender_vk: bytes, sbc):
        # Verify that the SubBlockContender message is validly constructured
        if not self._verify_sbc(sender_vk, sbc):
            self.log.error("Could not verify SBC from sender {}".format(sender_vk))
            return

        # If a sender 'resubmits' a SubBlockContender, it overwrites the previous. Seems like a potential attack vector?
        if sender_vk in self.sender_to_sbc:
            self.log.debug("Sender {} has already submitted a contender for sb idx {} with prev hash {}! Removing his "
                           "old contender before adding a new one".format(sender_vk, self.sb_idx, self.curr_block_hash))
            existing_sbc = self.sender_to_sbc[sender_vk]
            self.rh[existing_sbc.resultHash].remove(existing_sbc)

        # Add the SBC to the mapping of entries between SBCs and senders. This can be a set instead
        self.sender_to_sbc[sender_vk] = sbc

        # Add the SBC to the result hash. Shouldn't we hash it ourselves and just keep a hashmap?
        # Also, wouldn't it be easier to have a counter?
        self.rh[sbc.resultHash].add(sbc)

        # If the new subblock is now the 'best result', set it so in the instance.
        if (self.best_rh is None) or (len(self.rh[sbc.resultHash]) > len(self.rh[self.best_rh])):
            self.best_rh = sbc.resultHash

        # Not sure what this is doing
        for tx in sbc.transactions:

            h = hashlib.sha3_256()
            h.update(tx)
            _hash = h.digest()

            if _hash not in self.transactions:
                self.transactions[_hash] = tx

        self.log.info("Added SBC")

    def _verify_sbc(self, sender_vk: bytes, sbc) -> bool:
        # Dev sanity checks
        merkle_proof = Message.unpack_message_internal(MessageType.MERKLE_PROOF, sbc.signature)

        if not merkle_proof.signer == sender_vk:
            self.log.error('{} != {}'.format(merkle_proof.signer, sender_vk))
            return False

        if sbc.subBlockIdx != self.sb_idx:
            self.log.error('{} != {}'.format(sbc.subBlockIdx, self.sb_idx))
            return False

        # TODO move this validation to the SubBlockCotender objects instead
        # Validate signature

        valid_sig = _verify(vk=merkle_proof.signer,
                            msg=merkle_proof.hash,
                            signature=merkle_proof.signature)

        if not valid_sig:
            #self.log.error('SubBlockContender does not have a valid signature! SBC: {}'.format(sbc))
            return False

        # TODO move this validation to the SubBlockCotender objects instead
        # Validate sbc prev block hash matches our current block hash
        if sbc.prevBlockHash != self.curr_block_hash:
            self.log.error("SBC prev block hash {} does not match our current block hash {}!\nSBC: {}"
                           .format(sbc.prevBlockHash, self.curr_block_hash, sbc))
            # raghu todo - need to fix this
            return False

        # TODO move this validation to the SubBlockCotender objects instead
        # Validate merkle leaves
        if len(sbc.merkleLeaves) > 0:
            if not MerkleTree.verify_tree_from_bytes(leaves=sbc.merkleLeaves, root=sbc.resultHash):
                self.log.error("Could not verify MerkleTree for SBC {}!".format(sbc))
                return False

        # TODO move this validation to the SubBlockCotender objects instead
        # Validate transactions
        for tx in sbc.transactions:

            h = hashlib.sha3_256()
            h.update(tx)
            _hash = h.digest()

            if _hash not in sbc.merkleLeaves:
                self.log.error('Received malicious txs that does not match merkle leaves!\nSBC: {}'.format(sbc))
                return False

        # TODO move this validation to the SubBlockCotender objects instead
        # Validate sub block index is in range
        if sbc.subBlockIdx >= NUM_SB_PER_BLOCK:
            self.log.error("Got SBC with out of range sb_index {}!\nSBC: {}".format(sbc.subBlockIdx, sbc))
            return False

        return True


class BlockContender:
    def __init__(self):
        self.log = get_logger("BlockContender")
        self.committed = False
        self.consensus_reached = False
        self.state = MetaDataStorage()
        self.curr_block_hash = self.state.get_latest_block_hash()
        self.time_created = time.time()
        self.sb_groups = {}  # Mapping of sb indices to SubBlockGroup objects
        self.old_input_hashes = set()  # A set of input hashes from the last block.
        self.log.debug("BlockContender created with curr_block_hash={}".format(self.curr_block_hash))

    def reset(self):
        # Set old_input_hashes before we reset all the data
        # all_input_hashes = set()
        # for s in self._get_input_hashes():
            # all_input_hashes = all_input_hashes.union(s)
        # self.old_input_hashes = all_input_hashes
        # self.log.debugv("Old input hashes set to {}".format(self.old_input_hashes))

        # Reset all the data
        self.committed = False
        self.consensus_reached = False
        self.curr_block_hash = self.state.get_latest_block_hash()
        self.time_created = time.time()
        self.sb_groups = {}  # Mapping of sb indices to SubBlockGroup objects
        self.log.info("BlockContender reset with curr_block_hash={}".format(self.curr_block_hash))

    def is_consensus_reached(self) -> bool:
        assert len(self.sb_groups) <= NUM_SB_PER_BLOCK, "Got more sub block indices than SB_PER_BLOCK!"

        if len(self.sb_groups) != NUM_SB_PER_BLOCK:
            return False

        for sb_idx, sb_group in self.sb_groups.items():
            if not sb_group.is_consensus_reached():
                self.log.debugv("Consensus not reached yet on sb idx {}".format(sb_idx))
                return False

        return True

    def is_consensus_possible(self) -> bool:
        """
        Consensus is impossible if for any sub block contender the remaining votes were to go to the contender with the
        most votes and it still did not have 2/3 votes (i.e. contender with most votes + remaining votes < 2/3 votes)
        """
        for sb_idx, sb_group in self.sb_groups.items():
            if not sb_group.is_consensus_possible():
                return False

        return True

    def get_current_quorum_reached(self) -> int:
        if len(self.sb_groups) < NUM_SB_PER_BLOCK:
            return 0

        cur_quorum = PhoneBook.delegate_quorum_max
        for sb_idx, sb_group in self.sb_groups.items():
            cur_quorum = min(cur_quorum, sb_group.get_current_quorum_reached())

        return cur_quorum

    def is_empty(self):
        # assert self.is_consensus_reached(), "Consensus must be reached to check if this block is empty!"

        for sb_group in self.sb_groups.values():
            if not sb_group.is_empty():
                return False

        return True

    def get_sb_data(self):
        #assert self.is_consensus_reached(), "Cannot get block data if consensus is not reached!"
        #assert len(self.sb_groups) == NUM_SB_PER_BLOCK, "More sb_groups than subblocks! sb_groups={}".format(self.sb_groups)

        # Build the sub-blocks
        sb_data = []
        for sb_group in self.sb_groups.values():
            sb_data.append(sb_group.get_sb())

        sb_data = sorted(sb_data, key=lambda sb: sb.subBlockIdx)

        #assert len(sb_data) == NUM_SB_PER_BLOCK, "Block has {} sub blocks but there are {} SBs/per/block" \
                                                 #.format(len(sb_data), NUM_SB_PER_BLOCK)

        return sb_data

    def get_failed_block_notif(self):
        input_hashes = self._get_input_hashes()
        first_sb_idx = self._get_first_sb_idx()
        block_num = self.state.latest_block_num + 1
        last_hash = self.state.latest_block_hash
        return BlockNotification.get_failed_block_notification(
                                  block_num=block_num, prev_block_hash=last_hash,
                                  first_sb_idx=first_sb_idx, input_hashes=input_hashes)

    def add_sbc(self, sender_vk: str, sbc) -> bool:
        """
        Adds a SubBlockContender to this BlockContender's data.
        :param sender_vk: The VK of the sender of the SubBlockContender
        :param sbc: The SubBlockContender instance to add
        :return: True if this is the first SBC added to this BlockContender, and false otherwise
        """
        # Make sure this SBC does not refer to the last block created by checking the input hash
        if sbc.inputHash in self.old_input_hashes:
            self.log.info("Got SBC from prev block from sender {}! Ignoring.".format(sender_vk))  # TODO change log lvl?
            return False

        groups_empty = len(self.sb_groups) == 0

        if sbc.subBlockIdx not in self.sb_groups:
            self.sb_groups[sbc.subBlockIdx] = SubBlockGroup(sb_idx=sbc.subBlockIdx, curr_block_hash=self.curr_block_hash)

        self.sb_groups[sbc.subBlockIdx].add_sbc(sender_vk, sbc)
        return groups_empty

    # Return to this. Is this behaving properly? Is it redundant?
    def get_first_sb_idx_unsorted(self, sb_groups) -> int:
        sb_idx = sb_groups[0].sb_idx
        sbb_rem = sb_idx % NUM_SB_BUILDERS
        # assert sb_idx >= sbb_rem, "sub block indices are not maintained properly"
        return sb_idx - sbb_rem

    def get_first_sb_idx_sorted(self) -> int:
        sb_groups = sorted(self.sb_groups.values(), key=lambda sb: sb.sb_idx)
        num_sbg = len(sb_groups)
        assert num_sbg <= NUM_SB_PER_BLOCK, "Sub groups are not in a consistent state"
        return self.get_first_sb_idx_unsorted(sb_groups)

    # Pads empty sb groups?
    def get_input_hashes_sorted(self) -> List[list]:
        sb_groups = sorted(self.sb_groups.values(), key=lambda sb: sb.sb_idx)
        num_sbg = len(sb_groups)
        # assert num_sbg <= NUM_SB_PER_BLOCK, "Sub groups are not in a consistent state"

        sb_idx = self.get_first_sb_idx_unsorted(sb_groups) # 0
        input_hashes = []

        for sb_group in sb_groups:
            # In what natural situation does is while loop conditional met?
            while sb_idx < sb_group.sb_idx:
                sb_idx += 1
                num_sbg += 1
                input_hashes.append([])

            input_hashes.append(sb_group.get_input_hashes())
            sb_idx += 1

        # In what natural situation does is while loop conditional met?
        while num_sbg < NUM_SB_PER_BLOCK:
            num_sbg += 1
            input_hashes.append([])

        return input_hashes



import asyncio
from cilantro_ee.utils.test.testnet_config import set_testnet_config
set_testnet_config('2-2-2.json')

from cilantro_ee.core.logger.base import get_logger

from cilantro_ee.nodes.delegate.block_manager import BlockManager, IPC_PORT
from cilantro_ee.utils import int_to_bytes

from unittest import TestCase
from unittest import mock
from unittest.mock import MagicMock

from cilantro_ee.core.messages.message_type import MessageType
from cilantro_ee.core.messages.message import Message
from cilantro_ee.services.storage.vkbook import PhoneBook, VKBook

_log = get_logger("TestBlockManager")

from cilantro_ee.constants.testnet import TESTNET_DELEGATES, TESTNET_MASTERNODES
from cilantro_ee.services.storage.vkbook import VKBook
from cilantro_ee.constants.testnet import TESTNET_DELEGATES
TEST_IP = '127.0.0.1'
TEST_SK = TESTNET_DELEGATES[0]['sk']


class TestBlockManager(TestCase):

    # TODO we can probly DRY all this patching/setup code in a setup method or something
    @mock.patch("cilantro_ee.core.utils.worker.asyncio")
    @mock.patch("cilantro_ee.core.utils.worker.SocketManager")
    @mock.patch("cilantro_ee.nodes.delegate.block_manager.SubBlockBuilder")
    @mock.patch("cilantro_ee.nodes.delegate.block_manager.asyncio")
    @mock.patch("cilantro_ee.nodes.delegate.block_manager.BlockManager.run")
    def test_build_task_list_creates_ipc_router(self, mock_run_method, mock_bm_asyncio, mock_sbb, mock_manager, mock_worker_asyncio):
        PhoneBook = VKBook(delegates=[TESTNET_DELEGATES[0]['vk']], masternodes=[TESTNET_MASTERNODES[0]['vk']])
        bm = BlockManager(ip=TEST_IP, signing_key=TESTNET_DELEGATES[0]['sk'])

        # Need to set this manually to an empty list for tests only as overlay_client is mocked out
        bm.tasks = []

        # Attach a mock SockManager so create_socket calls return mock objects. Whenever a method is called on a mock
        # object, a _new mock object is automatically returned
        mock_manager = MagicMock()
        bm.manager = mock_manager

        # Mock manager.create_socket(...) to return a predictable list of mock objects that we can further assert on.
        # Relies on the sockets being created in EXACTLY the same order they are specified here in 'side_effect=[...]'
        mock_tcp_router = MagicMock()
        mock_ipc_router = MagicMock()
        mock_pub = MagicMock()
        mock_sub = MagicMock()
        mock_manager.create_socket = MagicMock(side_effect=[mock_tcp_router, mock_ipc_router, mock_pub, mock_sub])

        mock_router_handler_task = MagicMock()
        mock_tcp_router.add_handler = MagicMock(return_value=mock_router_handler_task)
        mock_ipc_router.add_handler = MagicMock(return_value=mock_router_handler_task)
        mock_pub.add_handler = MagicMock(return_value=mock_router_handler_task)
        mock_sub.add_handler = MagicMock(return_value=mock_router_handler_task)

        # Since we mocked out the .run method, we must invoke build_task_list manually
        bm.build_task_list()

        # Assert 'bind' was called on ipc_router with the expected args
        self.assertEqual(bm.ipc_router, mock_ipc_router)
        self.assertEqual(bm.router, mock_tcp_router)
        mock_ipc_router.bind.assert_called_with(port=IPC_PORT, protocol='ipc', ip=bm.ipc_ip)
        mock_ipc_router.add_handler.assert_called()
        self.assertTrue(mock_router_handler_task in bm.tasks)

    @mock.patch("cilantro_ee.core.utils.worker.asyncio")
    @mock.patch("cilantro_ee.core.utils.worker.SocketManager")
    @mock.patch("cilantro_ee.nodes.delegate.block_manager.SubBlockBuilder")
    @mock.patch("cilantro_ee.nodes.delegate.block_manager.asyncio")
    @mock.patch("cilantro_ee.nodes.delegate.block_manager.BlockManager.run")
    def test_handle_ready_msg(self, mock_run_method, mock_bm_asyncio, mock_sbb, mock_manager, mock_worker_asyncio):
        PhoneBook = VKBook(delegates=[TESTNET_DELEGATES[0]['vk']], masternodes=[TESTNET_MASTERNODES[0]['vk']])
        bm = BlockManager(ip=TEST_IP, signing_key=TESTNET_DELEGATES[0]['sk'])
        bm.tasks = []
        bm.manager = MagicMock()
        bm.ipc_router = MagicMock()
        bm._handle_sbc = MagicMock()

        bm.start_sbb_procs()

        num_sbb = len(bm.sb_builders)
        mtype, message = Message.get_message_packed(MessageType.READY)

        futs = []
        for i in range(num_sbb):
            self.assertFalse(bm.is_sbb_ready())
            frames = [str(i).encode(), mtype, message]
            futs.append(asyncio.ensure_future(bm.handle_ipc_msg(frames)))

        while len(futs) > 0:
            fut = futs.pop(0)
            if not fut.done():
                futs.append(fut)
            
        self.assertTrue(bm.is_sbb_ready())


    # TODO comment this back in when catchup is fixed
    # @mock.patch("cilantro_ee.core.utils.worker.asyncio", autospec=True)
    # @mock.patch("cilantro_ee.core.utils.worker.SocketManager", autospec=True)
    # @mock.patch("cilantro_ee.nodes.delegate.block_manager.SubBlockBuilder", autospec=True)
    # @mock.patch("cilantro_ee.nodes.delegate.block_manager.asyncio", autospec=True)
    # @mock.patch("cilantro_ee.nodes.delegate.block_manager.BlockManager.run", autospec=True)
    # def test_sub_msg_with_new_block_notification_calls_handle_new_block(self, mock_run_method, mock_bm_asyncio,
    #                                                                     mock_sbb, mock_manager, mock_worker_asyncio):
    #     """
    #     Tests handle_sub_msg correctly calls handle_new_block when a NewBlockNotification is received
    #     """
    #     bm = BlockManager(ip=TEST_IP, signing_key=TEST_SK)
    #     bm.manager = MagicMock()
    #     bm.handle_new_block = MagicMock()  # Mock out .handle_new_block
    #     bm.build_task_list()
    #
    #     # Mock Envelope.from_bytes to return a mock envelope of our choosing
    #     mock_env = MagicMock()
    #     mock_block_notif = MagicMock(spec=NewBlockNotification)
    #     fake_hash = 'DEADBEEF' * 8
    #     mock_env.message = mock_block_notif
    #     mock_env.message_hash = fake_hash
    #
    #     with mock.patch.object(Envelope, 'from_bytes', return_value=mock_env):
    #         # It doesnt actually matter what we pass in to bm.handle_sub_msg, since we've fixed Envelope.from_bytes
    #         # to return mock_env
    #         bm.handle_sub_msg([b'filter doesnt matter', b'envelope binary also doesnt matter'])
    #
    #     # Now, actually assert handle_new_block was called with mock_env as an arg
    #     bm.handle_new_block.assert_called_with(mock_env.message)


if __name__ == "__main__":
    import unittest
    unittest.main()
